
from StringIO import StringIO
import antlr

from stringtemplate3.utils import deprecated
from stringtemplate3.language.Expr import Expr
from stringtemplate3.language import ActionEvaluator
from stringtemplate3.language.StringTemplateAST import StringTemplateAST
from stringtemplate3.language.CatIterator import CatList
from stringtemplate3.language.FormalArgument import UNKNOWN_ARGS
import stringtemplate3

class IllegalStateException(Exception):

    def __init__(self, *args):
        Exception.__init__(self, *args)


def isiterable(o):
    if isinstance(o, (basestring, stringtemplate3.StringTemplate)):
        # don't consider strings and templates as iterables
        return False
    
    try:
        iter(o)
    except TypeError:
        return False
    else:
        return True
    
def convertAnyCollectionToList(o):
    list_ = None
    if isinstance(o, list):
        list_ = o
    elif isinstance(o, tuple) or isinstance(o, set):
        list_ = list(o)
    elif isinstance(o, dict):
        list_ = o.values()
    elif isinstance(o, CatList):
        list_ = []
        for item in o.lists():
            list_.append(item)
    if not list_:
        return o
    return list_


def convertAnythingToList(o):
    list_ = None
    if isinstance(o, list):
        list_ = o
    elif isinstance(o, tuple) or isinstance(o, set):
        list_ = list(o)
    elif isinstance(o, dict):
        list_ = o.values()
    elif isinstance(o, CatList):
        list_ = []
        for item in o.lists():
            list_.append(item)
    if not list_:
        return [o]
    return list_


class ASTExpr(Expr):
    """
    A single string template expression enclosed in $...; separator=...$
    parsed into an AST chunk to be evaluated.
    """

    DEFAULT_ATTRIBUTE_NAME = 'it'
    DEFAULT_ATTRIBUTE_KEY = 'ik'
    DEFAULT_ATTRIBUTE_NAME_DEPRECATED = 'attr'
    DEFAULT_INDEX_VARIABLE_NAME = 'i'
    DEFAULT_INDEX0_VARIABLE_NAME = 'i0'
    DEFAULT_MAP_VALUE_NAME = '_default_'
    DEFAULT_MAP_KEY_NAME = 'key'

    MAP_KEY_VALUE = None

    ## Using an expr option w/o value, makes options table hold EMPTY_OPTION
    #  value for that key.
    EMPTY_OPTION = "empty expr option"

    defaultOptionValues = {
        "anchor": StringTemplateAST(ActionEvaluator.STRING, "true"),
        "wrap": StringTemplateAST(ActionEvaluator.STRING, "\n")
        }

    supportedOptions = set([
        "anchor",
        "format",
        "null",
        "separator",
        "wrap"
        ])
    
    def __init__(self, enclosingTemplate, exprTree, options):
        super(ASTExpr, self).__init__(enclosingTemplate)
        self.exprTree = exprTree

        ## store separator etc...
        self.options = options

        ## A cached value of wrap=expr from the <...> expression.
        #  Computed in write(StringTemplate, StringTemplateWriter) and used
        #  in writeAttribute.
        self.wrapString = None

        ## For null values in iterated attributes and single attributes that
        #  are null, use this value instead of skipping.  For single valued
        #  attributes like <name; null="n/a"> it's a shorthand for
        #  <if(name)><name><else>n/a<endif>
        #  For iterated values <values; null="0", separator=",">, you get 0 for
        #  for null list values.  Works for template application like:
        #  <values:{v| <v>}; null="0"> also.
        self.nullValue = None

        ## A cached value of separator=expr from the <...> expression.
        #  Computed in write(StringTemplate, StringTemplateWriter) and used
        #  in writeAttribute.
        self.separatorString = None

        # A cached value of option format=expr
        self.formatString = None

        
    ## Return the tree interpreted when self template is written out.
    @deprecated
    def getAST(self):
        return self.exprTree

    def __str__(self):
        return str(self.exprTree)
    
    ## To write out the value of an ASTExpr, invoke the evaluator in eval.g
    #  to walk the tree writing out the values.  For efficiency, don't
    #  compute a bunch of strings and then pack them together.  Write out
    #  directly.
    #
    #  Compute separator and wrap expressions, save as strings so we don't
    #  recompute for each value in a multi-valued attribute or expression.
    #
    #  If they set anchor option, then inform the writer to push current
    #  char position.
    def write(self, this, out):
        if not self.exprTree or not this or not out:
            return 0
        out.pushIndentation(self.indentation)

        # handle options, anchor, wrap, separator...
        anchorAST = self.getOption("anchor")
        if anchorAST is not None: # any non-empty expr means true; check presence
            out.pushAnchorPoint()

        self.handleExprOptions(this)
        
        evaluator = ActionEvaluator.Walker()
        evaluator.initialize(this, self, out)
        n = 0
        try:
            # eval and write out tree
            n = evaluator.action(self.exprTree)
        except antlr.RecognitionException, re:
            this.error('can\'t evaluate tree: ' + self.exprTree.toStringList(),
                       re)
        out.popIndentation()
        if anchorAST is not None:
            out.popAnchorPoint()
        return n


    def handleExprOptions(self, this):
        """Grab and cache options; verify options are valid"""

        # make sure options don't use format / renderer.  They are usually
         # strings which might invoke a string renderer etc...
        self.formatString = None
        
        wrapAST = self.getOption("wrap")
        if wrapAST is not None:
            self.wrapString = self.evaluateExpression(this, wrapAST)

        nullValueAST = self.getOption("null")
        if nullValueAST is not None:
            self.nullValue = self.evaluateExpression(this, nullValueAST)

        separatorAST = self.getOption("separator")
        if separatorAST is not None:
            self.separatorString = self.evaluateExpression(this, separatorAST)

        formatAST = self.getOption("format")
        if formatAST is not None:
            self.formatString = self.evaluateExpression(this, formatAST)

        if self.options is not None:
            for option in self.options.keys():
                if option not in self.supportedOptions:
                    this.warning("ignoring unsupported option: "+option)
                    

# -----------------------------------------------------------------------------
#             HELP ROUTINES CALLED BY EVALUATOR TREE WALKER
# -----------------------------------------------------------------------------

    ## For <names,phones:{n,p | ...}> treat the names, phones as lists
    #  to be walked in lock step as n=names[i], p=phones[i].
    #
    def applyTemplateToListOfAttributes(self, this, attributes, templateToApply):
        if (not attributes) or (not templateToApply):
            return None # do not apply if missing templates or empty values
        argumentContext = None
        
        # indicate it's an ST-created list
        results = stringtemplate3.STAttributeList()

        # convert all attributes to iterators even if just one value
        attributesList = []
        for o in attributes:
            if o is not None:
                o = convertAnythingToList(o)
                attributesList.append(o)
        attributes = attributesList

        numAttributes = len(attributesList)

        # ensure arguments line up
        formalArgumentNames = templateToApply.formalArgumentKeys
        if not formalArgumentNames:
            this.error('missing arguments in anonymous template in context ' +
                       this.enclosingInstanceStackString)
            return None
        if len(formalArgumentNames) != numAttributes:
            this.error('number of arguments ' + str(formalArgumentNames) +
                       ' mismatch between attribute list and anonymous' +
                       ' template in context ' +
                       this.enclosingInstanceStackString)
            # truncate arg list to match smaller size
            shorterSize = min(len(formalArgumentNames), numAttributes)
            numAttributes = shorterSize
            formalArgumentNames = formalArgumentNames[:shorterSize]

        # keep walking while at least one attribute has values
        i = 0 # iteration number from 0
        while True:
            argumentContext = {}
            # get a value for each attribute in list; put into arg context
            # to simulate template invocation of anonymous template
            numEmpty = 0
            for a in range(numAttributes):
                if attributes[a]:
                    argName = formalArgumentNames[a]
                    argumentContext[argName] = attributes[a][0]
                    attributes[a] = attributes[a][1:]
                else:
                    numEmpty += 1
            if numEmpty == numAttributes: # No attribute values given
                break
            argumentContext[self.DEFAULT_INDEX_VARIABLE_NAME] = i + 1
            argumentContext[self.DEFAULT_INDEX0_VARIABLE_NAME] = i
            embedded = templateToApply.getInstanceOf()
            embedded.enclosingInstance = this
            embedded.argumentContext = argumentContext
            results.append(embedded)
            i += 1
            
        return results

    def applyListOfAlternatingTemplates(self, this, attributeValue, templatesToApply):
        if not attributeValue or not templatesToApply or templatesToApply == []:
            # do not apply if missing templates or empty value
            return None

        embedded = None
        argumentContext = None

        if isiterable(attributeValue):
            # results can be treated list an attribute, indicate ST created list
            resultVector = stringtemplate3.STAttributeList()
            for i, ithValue in enumerate(attributeValue):
                if ithValue is None:
                    if self.nullValue is None:
                        continue
                    ithValue = self.nullValue

                templateIndex = i % len(templatesToApply) # rotate through
                embedded = templatesToApply[templateIndex]
                # template to apply is an actual stringtemplate.StringTemplate (created in
                # eval.g), but that is used as the examplar.  We must create
                # a new instance of the embedded template to apply each time
                # to get new attribute sets etc...
                args = embedded.argumentsAST
                embedded = embedded.getInstanceOf() # make new instance
                embedded.enclosingInstance = this
                embedded.argumentsAST = args
                argumentContext = {}
                formalArgs = embedded.formalArguments
                isAnonymous = embedded.name == stringtemplate3.ANONYMOUS_ST_NAME
                self.setSoleFormalArgumentToIthValue(embedded, argumentContext, ithValue)
                if isinstance(attributeValue, dict):
                    argumentContext[self.DEFAULT_ATTRIBUTE_KEY] = ithValue
                    # formalArgs might be UNKNOWN, which is non-empty but treated
                    # like 'no formal args'. FIXME: Is this correct
                    if not (isAnonymous and formalArgs is not None and len(formalArgs) > 0 and formalArgs != UNKNOWN_ARGS):
                        argumentContext[self.DEFAULT_ATTRIBUTE_NAME] = { ithValue: attributeValue[ithValue] }
                        argumentContext[self.DEFAULT_ATTRIBUTE_NAME_DEPRECATED] = { ithValue: attributeValue[ithValue] }
                else:
                    argumentContext[self.DEFAULT_ATTRIBUTE_KEY] = None
                    # formalArgs might be UNKNOWN, which is non-empty but treated
                    # like 'no formal args'. FIXME: Is this correct
                    if not (isAnonymous and formalArgs is not None and len(formalArgs) > 0 and formalArgs != UNKNOWN_ARGS):
                        argumentContext[self.DEFAULT_ATTRIBUTE_NAME] = ithValue
                        argumentContext[self.DEFAULT_ATTRIBUTE_NAME_DEPRECATED] = ithValue
                argumentContext[self.DEFAULT_INDEX_VARIABLE_NAME] = i+1
                argumentContext[self.DEFAULT_INDEX0_VARIABLE_NAME] = i
                embedded.argumentContext = argumentContext
                self.evaluateArguments(embedded)
                #sys.stderr.write('i=' + str(i) + ': applyTemplate(' +
                #                 embedded.getName() + ', args=' +
                #                 str(argumentContext) + ' to attribute value ' +
                #                 str(ithValue) + '\n')
                resultVector.append(embedded)

            if not len(resultVector):
                resultVector = None

            return resultVector

        else:
            embedded = templatesToApply[0]
            #sys.stderr.write('setting attribute ' +
            #                 DEFAULT_ATTRIBUTE_NAME +
            #                 ' in arg context of ' + embedded.getName() +
            #                 ' to ' + str(attributeValue) + '\n')
            argumentContext = {}
            formalArgs = embedded.formalArguments
            args = embedded.argumentsAST
            self.setSoleFormalArgumentToIthValue(embedded, argumentContext, attributeValue)
            isAnonymous = embedded.name == stringtemplate3.ANONYMOUS_ST_NAME
            # if it's an anonymous template with a formal arg, don't set it/attr
            # formalArgs might be UNKNOWN, which is non-empty but treated
            # like 'no formal args'. FIXME: Is this correct
            if not (isAnonymous and formalArgs is not None and len(formalArgs) > 0 and formalArgs != UNKNOWN_ARGS):
                argumentContext[self.DEFAULT_ATTRIBUTE_NAME] = attributeValue
                argumentContext[self.DEFAULT_ATTRIBUTE_NAME_DEPRECATED] = attributeValue
            argumentContext[self.DEFAULT_INDEX_VARIABLE_NAME] = 1
            argumentContext[self.DEFAULT_INDEX0_VARIABLE_NAME] = 0
            embedded.argumentContext = argumentContext
            self.evaluateArguments(embedded)
            return embedded

    def setSoleFormalArgumentToIthValue(self, embedded, argumentContext, ithValue):
        formalArgs = embedded.formalArguments
        if formalArgs:
            soleArgName = None
            isAnonymous = (embedded.name == stringtemplate3.ANONYMOUS_ST_NAME)
            if len(formalArgs) == 1 or (isAnonymous and len(formalArgs) > 0):
                if isAnonymous and len(formalArgs) > 1:
                    embedded.error('too many arguments on {...} template: ' +
                                   str(formalArgs))
                # if exactly 1 arg or anonymous, give that the value of
                # "it" as a convenience like they said
                # $list:template(arg=it)$
                argNames = formalArgs.keys()
                soleArgName = argNames[0]
                argumentContext[soleArgName] = ithValue

    ## Return o.getPropertyName() given o and propertyName.  If o is
    #  a stringtemplate then access it's attributes looking for propertyName
    #  instead (don't check any of the enclosing scopes; look directly into
    #  that object).  Also try isXXX() for booleans.  Allow HashMap,
    #  Hashtable as special case (grab value for key).
    #
    def getObjectProperty(self, this, o, propertyName):
        if (not o) or (not propertyName):
            return None
        value = None

        # Special case: our automatically created Aggregates via
        # attribute name: "{obj.{prop1,prop2}}"
        if isinstance(o, stringtemplate3.Aggregate):
            #sys.stderr.write("[getObjectProperty] o = " + str(o) + '\n')
            value = o.get(propertyName, None)
            if value is None:
                # no property defined; if a map in this group
                # then there may be a default value
                value = o.get(self.DEFAULT_MAP_VALUE_NAME, None)

        # Or: if it's a dictionary then pull using key not the
        # property method.
        elif isinstance(o, dict):
            if propertyName == 'keys':
                value = o.keys()

            elif propertyName == 'values':
                value = o.values()

            else:
                value = o.get(propertyName, None)
                if value is None:
                    value = o.get(self.DEFAULT_MAP_VALUE_NAME, None)

            if value is self.MAP_KEY_VALUE:
                value = propertyName
                
            return value

        # Special case: if it's a template, pull property from
        # it's attribute table.
        # TODO: TJP just asked himself why we can't do inherited attr here?
        elif isinstance(o, stringtemplate3.StringTemplate):
            attributes = o.attributes
            if attributes:
                if attributes.has_key(propertyName): # prevent KeyError...
                    value = attributes[propertyName]
                else:
                    value = None

        else:
            # use getPropertyName() lookup
            methodSuffix = propertyName[0].upper() + propertyName[1:]
            methodName = 'get' + methodSuffix
            m = None
            if hasattr(o, methodName):
                m = getattr(o, methodName)
            else:
                methodName = 'is' + methodSuffix
                try:
                    m = getattr(o, methodName)
                except AttributeError, ae:
                    # try for a visible field
                    try:
                        try:
                            return getattr(o, propertyName)
                        except AttributeError, ae2:
                            this.error('Can\'t get property ' + propertyName +
                                       ' using method get/is' + methodSuffix +
                                       ' or direct field access from ' +
                                       o.__class__.__name__ + ' instance', ae2)
                    except AttributeError, ae:
                        this.error('Class ' + o.__class__.__name__ +
                                   ' has no such attribute: ' + propertyName +
                                   ' in template context ' +
                                   this.enclosingInstanceStackString, ae)

            if m is not None:
                try:
                    value = m()
                except Exception, e:
                    this.error('Can\'t get property ' + propertyName +
                               ' using method get/is' + methodSuffix +
                               ' or direct field access from ' +
                               o.__class__.__name__ + ' instance', e)

        return value

    ## Normally StringTemplate tests presence or absence of attributes
    #  for adherence to my principles of separation, but some people
    #  disagree and want to change.
    #
    #  For 2.0, if the object is a boolean, do something special. $if(boolean)$
    #  will actually test the value.  Now, self breaks my rules of entanglement
    #  listed in my paper, but it truly surprises programmers to have booleans
    #  always True.  Further, the key to isolating logic in the model is avoiding
    #  operators (for which you need attribute values).  But, no operator is
    #  needed to use boolean values.  Well, actually I guess "!" (not) is
    #  an operator.  Regardless, for practical reasons, I'm going to technically
    #  violate my rules as I currently have them defined.  Perhaps for a future
    #  version of the paper I will refine the rules.
    #
    #  Post 2.1, I added a check for non-null Iterators, Collections, ...
    #  with size==0 to return false. TJP 5/1/2005
    #
    def testAttributeTrue(self, a):
        if not a:
            return False
        if isinstance(a, bool):
            return a
        return True


    ## For now, we can only add two objects as strings; convert objects to
    #  strings then cat.
    #
    def add(self, a, b):
        # a None value means don't do cat, just return other value
        if a is None:
            return b
        elif b is None:
            return a

        return unicode(a) + unicode(b)
    
    ## Call a string template with args and return result.  Do not convert
    #  to a string yet.  It may need attributes that will be available after
    #  self is inserted into another template.
    #
    def getTemplateInclude(self, enclosing, templateName, argumentsAST):
        group = enclosing.group
        embedded = group.getEmbeddedInstanceOf(templateName, enclosing)
        if not embedded:
            enclosing.error('cannot make embedded instance of ' +
                            templateName + ' in template ' +
                            enclosing.getName())
            return None

        embedded.argumentsAST = argumentsAST
        self.evaluateArguments(embedded)
        return embedded

    ## How to spit out an object.  If it's not a StringTemplate nor a sequence,
    #  just do o.toString().  If it's a StringTemplate, do o.write(out).
    #  If it's a sequence, do a write(out, o[i]) for all elements.
    #  Note that if you do something weird like set the values of a
    #  multivalued tag to be sequences, it will effectively flatten it.
    #
    #  If this is an embedded template, you might have specified
    #  a separator arg; used when is a sequence.
    #
    def writeAttribute(self, this, o, out):
        return self._write(this, o, out)

    def _write(self, this, o, out):
        """
        Write o relative to self to out.

        John Snyders fixes here for formatString.  Basically, any time
        you are about to write a value, check formatting.
        """

        if o is None:
            if self.nullValue is None:
                return 0
            o = self.nullValue
            
        n = 0
        try:
            if isinstance(o, stringtemplate3.StringTemplate):
                # failsafe: perhaps enclosing instance not set
                # Or, it could be set to another context!  This occurs
                # when you store a template instance as an attribute of more
                # than one template (like both a header file and C file when
                # generating C code).  It must execute within the context of
                # the enclosing template.
                o.enclosingInstance = this
                # if this is found up the enclosing instance chain, then
                # infinite recursion
                if stringtemplate3.lintMode and \
                   stringtemplate3.StringTemplate.isRecursiveEnclosingInstance(o):
                    # throw exception since sometimes eval keeps going
                    # even after I ignore self write of o.
                    raise IllegalStateException('infinite recursion to ' +
                        o.templateDeclaratorString + ' referenced in ' +
                        o.enclosingInstance.templateDeclaratorString +
                        '; stack trace:\n' + o.enclosingInstanceStackTrace)
                else:
                    # if we have a wrap string, then inform writer it
                    # might need to wrap
                    if self.wrapString is not None:
                        n = out.writeWrapSeparator(self.wrapString)

                    # check if formatting needs to be applied to the stToWrite
                    if self.formatString is not None:
                        renderer = this.getAttributeRenderer(str)
                        if renderer is not None:
                            # you pay a penalty for applying format option to a
                            # template because the template must be written to
                            # a temp StringWriter so it can be formatted before
                            # being written to the real output.
                            buf = StringIO
                            sw = this.getGroup().getStringTemplateWriter(buf)
                            o.write(sw)
                            n = out.write(renderer.toString(buf.getvalue(), self.formatString))
                            return n
                        
                    n = o.write(out)
                return n

            if isiterable(o):
                if isinstance(o, dict):
                    # for mapping we want to iterate over the values
                    lst = o.values()
                else:
                    lst = o

                seenPrevValue = False
                for iterValue in lst:
                    if iterValue is None:
                        iterValue = self.nullValue

                    if iterValue is not None:
                        if ( seenPrevValue and
                             self.separatorString is not None ):
                            n += out.writeSeparator(self.separatorString)
                            
                        seenPrevValue = True
                        n += self._write(this, iterValue, out)
                        

            else:
                renderer = this.getAttributeRenderer(o.__class__)
                if renderer is not None:
                    v = renderer.toString(o, self.formatString)
                else:
                    v = unicode(o)

                if self.wrapString is not None:
                    n = out.write(v, self.wrapString)
                else:
                    n = out.write(v)
                    
                return n
            
        except IOError, io:
            this.error('problem writing object: ' + o, io)
        return n


    def evaluateExpression(self, this, expr):
        """
        A expr is normally just a string literal, but is still an AST that
        we must evaluate.  The expr can be any expression such as a template
        include or string cat expression etc...  Evaluate with its own writer
        so that we can convert to string and then reuse, don't want to compute
        all the time; must precompute w/o writing to output buffer.
        """
        
        if expr is None:
            return None

        if isinstance(expr, StringTemplateAST):
            # must evaluate, writing to a string so we can hang on to it
            buf = StringIO()
            
            sw = this.group.getStringTemplateWriter(buf)
            evaluator = ActionEvaluator.Walker()
            evaluator.initialize(this, self, sw)
            try:
                evaluator.action(expr) # eval tree

            except antlr.RecognitionException, re:
                this.error(
                    "can't evaluate tree: "+self.exprTree.toStringList(), re
                    )

            return buf.getvalue()

        else:
            # just in case we expand in the future and it's something else
            return str(expr)


    ## Evaluate an argument list within the context of the enclosing
    #  template but store the values in the context of self, the
    #  new embedded template.  For example, bold(item=item) means
    #  that bold.item should get the value of enclosing.item.
    #
    def evaluateArguments(self, this):
        argumentsAST = this.argumentsAST
        if not argumentsAST or not argumentsAST.getFirstChild():
            # return immediately if missing tree or no actual args
            return

        # Evaluate args in the context of the enclosing template, but we
        # need the predefined args like 'it', 'attr', and 'i' to be
        # available as well so we put a dummy ST between the enclosing
        # context and the embedded context.  The dummy has the predefined
        # context as does the embedded.
        enclosing = this.enclosingInstance
        argContextST = stringtemplate3.StringTemplate(
            group=this.group,
            template=""
            )
        argContextST.name = '<invoke ' + this.name + ' arg context>'
        argContextST.enclosingInstance = enclosing
        argContextST.argumentContext = this.argumentContext

        eval_ = ActionEvaluator.Walker()
        eval_.initialize(argContextST, self, None)
        #sys.stderr.write('eval args: ' + argumentsAST.toStringList() + '\n')
        #sys.stderr.write('ctx is ' + this.getArgumentContext())
        try:
            # using any initial argument context (such as when obj is set),
            # evaluate the arg list like bold(item=obj).  Since we pass
            # in any existing arg context, that context gets filled with
            # new values.  With bold(item=obj), context becomes:
            #:[obj=...],[item=...]}.
            ac = eval_.argList(argumentsAST, this, this.argumentContext)
            this.argumentContext = ac

        except antlr.RecognitionException, re:
            this.error('can\'t evaluate tree: ' + argumentsAST.toStringList(),
                       re)

    ## Return the first attribute if multiple valued or the attribute
    #  itself if single-valued.  Used in <names:first()>
    #
    def first(self, attribute):
        if not attribute:
            return None
        f = attribute
        attribute = convertAnyCollectionToList(attribute)
        if attribute and isinstance(attribute, list):
            f = attribute[0]
        return f

    ## Return the everything but the first attribute if multiple valued
    #  or null if single-valued.  Used in <names:rest()>.
    #
    def rest(self, attribute):
        if not attribute:
            return None
        theRest = attribute
        attribute = convertAnyCollectionToList(attribute)
        if not attribute:
            # if not even one value return None
            return None
        if isinstance(attribute, list):
            # ignore first value
            attribute = attribute[1:]
            if not attribute:
                # if not more than one value, return None
                return None
            # return suitably altered iterator
            theRest = attribute
        else:
            # rest of single-valued attribute is None
            theRest = None
        return theRest

    ## Return the last attribute if multiple valued or the attribute
    #  itself if single-valued.  Used in <names:last()>.  This is pretty
    #  slow as it iterates until the last element.  Ultimately, I could
    #  make a special case for a List or Vector.
    #
    def last(self, attribute):
        if not attribute:
            return None
        l = attribute
        attribute = convertAnyCollectionToList(attribute)
        if attribute and isinstance(attribute, list):
            l = attribute[-1]
        return l


    def strip(self, attribute):
        """Return an iterator that skips all null values."""

        if attribute is None:
            yield None

        elif isinstance(attribute, basestring):
            # don't iterate over string
            yield attribute

        else:
            try:
                it = iter(attribute)
            except TypeError:
                # attribute is not iterable
                yield attribute

            else:
                for value in it:
                    if value is not None:
                        yield value
                

    def trunc(self, attribute):
        """
        Return all but the last element.  trunc(x)=null if x is single-valued.
        """
        return None # not impl.


    def length(self, attribute):
        """
        Return the length of a multiple valued attribute or 1 if it is a
        single attribute. If attribute is null return 0.
        Special case several common collections and primitive arrays for
        speed.  This method by Kay Roepke.
        """

        if attribute is None:
            return 0

        if isinstance(attribute, (dict, list)):
            i = len(attribute)

        elif isinstance(attribute, basestring):
            # treat strings as atoms
            i = 1

        else:
            try:
                it = iter(attribute)
                
            except TypeError:
                # can't iterate over it, we have at least one of something.
                i = 1
                
            else:
                # count items
                i = sum(1 for _ in it)

        return i


    def getOption(self, name):
        value = None
        if self.options is not None:
            value = self.options.get(name, None)
            if value == self.EMPTY_OPTION:
                return self.defaultOptionValues.get(name, None)

        return value





# [The "BSD licence"]
# Copyright (c) 2003-2006 Terence Parr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import sys
import traceback
import imp
import time
from StringIO import StringIO

import antlr

from stringtemplate3.language import (
    AngleBracketTemplateLexer,
    DefaultTemplateLexer,
    GroupLexer, GroupParser,
    )

from stringtemplate3.utils import deprecated
from stringtemplate3.errors import (
    DEFAULT_ERROR_LISTENER
    )
from stringtemplate3.templates import (
    StringTemplate, REGION_IMPLICIT
    )
from stringtemplate3.writers import AutoIndentWriter
from stringtemplate3.interfaces import StringTemplateGroupInterface

DEFAULT_EXTENSION = '.st'

## Used to indicate that the template doesn't exist.
#  We don't have to check disk for it; we know it's not there.
#  Set later to work around cyclic class definitions
NOT_FOUND_ST = None


class StringTemplateGroup(object):
    """
    Manages a group of named mutually-referential StringTemplate objects.
    Currently the templates must all live under a directory so that you
    can reference them as foo.st or gutter/header.st.  To refresh a
    group of templates, just create a new StringTemplateGroup and start
    pulling templates from there.  Or, set the refresh interval.

    Use getInstanceOf(template-name) to get a string template
    to fill in.

    The name of a template is the file name minus ".st" ending if present
    unless you name it as you load it.

    You can use the group file format also to define a group of templates
    (this works better for code gen than for html page gen).  You must give
    a Reader to the ctor for it to load the group; this is general and
    distinguishes it from the ctors for the old-style "load template files
    from the disk".

    10/2005 I am adding a StringTemplateGroupLoader concept so people can
    define supergroups within a group and have it load that group
    automatically.
    """
    
    ## Track all groups by name; maps name to StringTemplateGroup
    nameToGroupMap = {}

    ## Track all interfaces by name; maps name to StringTemplateGroupInterface
    nameToInterfaceMap = {}
    
    ## If a group file indicates it derives from a supergroup, how do we
    #  find it?  Shall we make it so the initial StringTemplateGroup file
    #  can be loaded via this loader?  Right now we pass a Reader to ctor
    #  to distinguish from the other variety.
    groupLoader = None

    ## You can set the lexer once if you know all of your groups use the
    #  same separator.  If the instance has templateLexerClass set
    #  then it is used as an override.
    defaultTemplateLexerClass = DefaultTemplateLexer.Lexer
        
    def __init__(self, name=None, rootDir=None, lexer=None, file=None, errors=None, superGroup=None):
        ## What is the group name
        #
        self.name = None
        ## Maps template name to StringTemplate object
        #
        self.templates = {}
        ## Maps map names to HashMap objects.  This is the list of maps
        #  defined by the user like typeInitMap ::= ["int":"0"]
        #
        self.maps = {}

        ## How to pull apart a template into chunks?
        self._templateLexerClass = None
        
        ## Under what directory should I look for templates?  If None,
        #  to look into the CLASSPATH for templates as resources.
        #
        self.rootDir = None

        ## Are we derived from another group?  Templates not found in this
        #  group will be searched for in the superGroup recursively.
        self._superGroup = None

        ## Keep track of all interfaces implemented by this group.
        self.interfaces = []
        
        ## When templates are files on the disk, the refresh interval is used
        #  to know when to reload.  When a Reader is passed to the ctor,
        #  it is a stream full of template definitions.  The former is used
        #  for web development, but the latter is most likely used for source
        #  code generation for translators; a refresh is unlikely.  Anyway,
        #  I decided to track the source of templates in case such info is useful
        #  in other situations than just turning off refresh interval.  I just
        #  found another: don't ever look on the disk for individual templates
        #  if this group is a group file...immediately look into any super group.
        #  If not in the super group, report no such template.
        #
        self.templatesDefinedInGroupFile = False
        ## Normally AutoIndentWriter is used to filter output, but user can
        #  specify a new one.
        #
        self.userSpecifiedWriter = None

        self.debugTemplateOutput = False

        ## The set of templates to ignore when dumping start/stop debug strings
        self.noDebugStartStopStrings = None
        
        ## A Map<class,object> that allows people to register a renderer for
        #  a particular kind of object to be displayed for any template in this
        #  group.  For example, a date should be formatted differently depending
        #  on the locale.  You can set Date.class to an object whose
        #  str() method properly formats a Date attribute
        #  according to locale.  Or you can have a different renderer object
        #  for each locale.
        #
        #  These render objects are used way down in the evaluation chain
        #  right before an attribute's str() method would normally be
        #  called in ASTExpr.write().
        #
        self.attributeRenderers = None

        ## Where to report errors.  All string templates in this group
        #  use this error handler by default.
        if errors is not None:
            self.listener = errors
        else:
            self.listener = DEFAULT_ERROR_LISTENER
            
        ## How long before tossing out all templates in seconds.
        #  default: no refreshing from disk
        #
        self.refreshInterval = sys.maxint/1000
        self.lastCheckedDisk = 0L
        
        if name is not None:
            assert isinstance(name, basestring)
            self.name = name

            assert rootDir is None or isinstance(rootDir, basestring)
            self.rootDir = rootDir
            self.lastCheckedDisk = time.time()
            StringTemplateGroup.nameToGroupMap[self.name] = self

            self.templateLexerClass = lexer
            
            assert superGroup is None or isinstance(superGroup, StringTemplateGroup)
            self.superGroup = superGroup


        if file is not None:
            assert hasattr(file, 'read')
            
            self.templatesDefinedInGroupFile = True

            if lexer is not None:
                self.templateLexerClass = lexer
            else:
                self.templateLexerClass = AngleBracketTemplateLexer.Lexer

            assert superGroup is None or isinstance(superGroup, StringTemplateGroup)
            self.superGroup = superGroup

            self.parseGroup(file)
            assert self.name is not None
            StringTemplateGroup.nameToGroupMap[self.name] = self
            self.verifyInterfaceImplementations()


    def getTemplateLexerClass(self):
        """
        What lexer class to use to break up templates.  If not lexer set
        for this group, use static default.
        """

        if self._templateLexerClass is not None:
            return self._templateLexerClass

        return self.defaultTemplateLexerClass

    def setTemplateLexerClass(self, lexer):
        if isinstance(lexer, basestring):
            try:
                self._templateLexerClass = {
                    'default': DefaultTemplateLexer.Lexer,
                    'angle-bracket': AngleBracketTemplateLexer.Lexer,
                    }[lexer]
            except KeyError:
                raise ValueError('Unknown lexer id %r' % lexer)

        elif isinstance(lexer, type) and issubclass(lexer, antlr.CharScanner):
            self._templateLexerClass = lexer

        elif lexer is not None:
            raise TypeError(
                "Lexer must be string or lexer class, got %r"
                % type(lexer).__name__
                )

    templateLexerClass = property(getTemplateLexerClass, setTemplateLexerClass)
    getTemplateLexerClass = deprecated(getTemplateLexerClass)
    setTemplateLexerClass = deprecated(setTemplateLexerClass)


    @deprecated
    def getName(self):
        return self.name

    @deprecated
    def setName(self, name):
        self.name = name


    def setSuperGroup(self, superGroup):
        if superGroup is None or isinstance(superGroup, StringTemplateGroup):
            self._superGroup = superGroup
            
        elif isinstance(superGroup, basestring):
            # Called by group parser when ": supergroupname" is found.
            # This method forces the supergroup's lexer to be same as lexer
            # for this (sub) group.

            superGroupName = superGroup
            superGroup = StringTemplateGroup.nameToGroupMap.get(
                superGroupName, None)
            if superGroup is not None:
                # we've seen before; just use it
                self.superGroup = superGroup

            else:
                # else load it using this group's template lexer
                superGroup = self.loadGroup(
                    superGroupName, lexer=self.templateLexerClass)
                if superGroup is not None:
                    StringTemplateGroup.nameToGroupMap[superGroup] = superGroup
                    self._superGroup = superGroup

                elif self.groupLoader is None:
                    self.listener.error("no group loader registered", None)

        else:
            raise TypeError(
                "Need StringTemplateGroup or string, got %s"
                % type(superGroup).__name__
                )

    def getSuperGroup(self):
        return self._superGroup

    superGroup = property(getSuperGroup, setSuperGroup)
    getSuperGroup = deprecated(getSuperGroup)
    setSuperGroup = deprecated(setSuperGroup)
    

    def getGroupHierarchyStackString(self):
        """Walk up group hierarchy and show top down to this group"""

        groupNames = []
        p = self
        while p is not None:
            groupNames.insert(0, p.name)
            p = p.superGroup

        return '[' + ' '.join(groupNames) + ']'


    @deprecated
    def getRootDir(self):
        return self.rootDir

    @deprecated
    def setRootDir(self, rootDir):
        self.rootDir = rootDir


    def implementInterface(self, interface):
        """
        Indicate that this group implements this interface.
        Load if necessary if not in the nameToInterfaceMap.
        """

        if isinstance(interface, StringTemplateGroupInterface):
            self.interfaces.append(interface)

        else:
            interfaceName = interface

            interface = self.nameToInterfaceMap.get(interfaceName, None)
            if interface is not None:
                # we've seen before; just use it
                self.interfaces.append(interface)
                return

            # else load it
            interface = self.loadInterface(interfaceName)
            if interface is not None:
                self.nameToInterfaceMap[interfaceName] = interface
                self.interfaces.append(interface)

            elif self.groupLoader is None:
                self.listener.error("no group loader registered", None)


    ## StringTemplate object factory; each group can have its own.
    def createStringTemplate(self):
        return StringTemplate()


    def getInstanceOf(self, name, enclosingInstance=None, attributes=None):
        """
        A support routine that gets an instance of name knowing which
        ST encloses it for error messages.
        """

        assert isinstance(name, basestring)
        assert enclosingInstance is None or isinstance(enclosingInstance, StringTemplate)
        assert attributes is None or isinstance(attributes, dict)

        st = self.lookupTemplate(name, enclosingInstance)
        if st is not None:
            st = st.getInstanceOf()
            
            if attributes is not None:
                st.attributes = attributes
                
            return st
        
        return None

    def getEmbeddedInstanceOf(self, name, enclosingInstance):
        assert isinstance(name, basestring)
        assert enclosingInstance is None or isinstance(enclosingInstance, StringTemplate)

        st = None
        # TODO: seems like this should go into lookupTemplate
        if name.startswith("super."):
            # for super.foo() refs, ensure that we look at the native
            # group for the embedded instance not the current evaluation
            # group (which is always pulled down to the original group
            # from which somebody did group.getInstanceOf("foo");
            st = enclosingInstance.nativeGroup.getInstanceOf(
                name, enclosingInstance
                )
            
        else:
            st = self.getInstanceOf(name, enclosingInstance)

        # make sure all embedded templates have the same group as enclosing
        # so that polymorphic refs will start looking at the original group
        st.group = self
        st.enclosingInstance = enclosingInstance
        return st

    ## Get the template called 'name' from the group.  If not found,
    #  attempt to load.  If not found on disk, then try the superGroup
    #  if any.  If not even there, then record that it's
    #  NOT_FOUND so we don't waste time looking again later.  If we've gone
    #  past refresh interval, flush and look again.
    #
    #  If I find a template in a super group, copy an instance down here
    def lookupTemplate(self, name, enclosingInstance=None):
        assert isinstance(name, basestring)
        assert enclosingInstance is None or isinstance(enclosingInstance, StringTemplate)

        if name.startswith('super.'):
            if self.superGroup:
                dot = name.find('.')
                name = name[dot+1:]
                return self.superGroup.lookupTemplate(name, enclosingInstance)
            raise ValueError(self.name + ' has no super group; ' +
                             'invalid template: ' + name)
        self.checkRefreshInterval()
        st = self.templates.get(name, None)
        if not st:
            # not there?  Attempt to load
            if not self.templatesDefinedInGroupFile:
                # only check the disk for individual template
                st = self.loadTemplateFromBeneathRootDir(self.getFileNameFromTemplateName(name))
            if (not st) and self.superGroup:
                # try to resolve in super group
                st = self.superGroup.getInstanceOf(name)
                # make sure that when we inherit a template, that it's
                # group is reset; it's nativeGroup will remain where it was
                if st is not None:
                    st.group = self

            if st: # found in superGroup
                # insert into this group; refresh will allow super
                # to change it's def later or this group to add
                # an override.
                self.templates[name] = st
                
            else:
                # not found; remember that this sucker doesn't exist
                self.templates[name] = StringTemplateGroup.NOT_FOUND_ST
                context = ""
                if enclosingInstance is not None:
                    context = (
                        "; context is "+
                        enclosingInstance.enclosingInstanceStackString
                        )
                hier = self.getGroupHierarchyStackString()
                context += "; group hierarchy is "+hier
                raise ValueError(
                    "Can't load template " +
                    self.getFileNameFromTemplateName(name) +
                    context
                    )

        elif st is StringTemplateGroup.NOT_FOUND_ST:
            return None

        return st

    def checkRefreshInterval(self):
        if self.templatesDefinedInGroupFile:
            return
        if self.refreshInterval == 0 or \
           (time.time() - self.lastCheckedDisk) >= \
           self.refreshInterval:
            # throw away all pre-compiled references
            self.templates = {}
            self.lastCheckedDisk = time.time()


    def loadTemplate(self, name, src):
        if isinstance(src, basestring):
            template = None
            try:
                br = open(src, 'r')
                try:
                    template = self.loadTemplate(name, br)
                finally:
                    br.close()

            # FIXME: eek, that's ugly
            except Exception, e:
                raise
            
            return template
        
        elif hasattr(src, 'readlines'):
            buf = src.readlines()
            
            # strip newlines etc.. from front/back since filesystem
            # may add newlines etc...
            pattern = str().join(buf).strip()
            
            if not pattern:
                self.error("no text in template '"+name+"'")
                return None
            
            return self.defineTemplate(name, pattern)
        
        raise TypeError(
            'loadTemplate should be called with a file or filename'
            )


    ## Load a template whose name is derived from the template filename.
    #  If there is a rootDir, try to load the file from there.
    #
    def loadTemplateFromBeneathRootDir(self, fileName):
        template = None
        name = self.getTemplateNameFromFileName(fileName)
        # if no rootDir, try to load as a resource in CLASSPATH
        # In the Python case that is of course the sys.path
        if not self.rootDir:
            try:
                br, pathName, descr = imp.find_module(name)
            except ImportError:
                br = None
            if br is None:
                return None

            try:
                try:
                    template = self.loadTemplate(name, br)
                except IOError, ioe:
                    self.error("Problem reading template file: "+fileName, ioe)

            finally:
                try:
                    br.close()
                except IOError, ioe2:
                    self.error('Cannot close template file: ' + pathName, ioe2)

            return template
        # load via rootDir
        template = self.loadTemplate(name, self.rootDir + '/' + fileName)
        return template

    ## (def that people can override behavior; not a general
    #  purpose method)
    #
    def getFileNameFromTemplateName(self, templateName):
        return templateName + DEFAULT_EXTENSION

    ## Convert a filename relativePath/name.st to relativePath/name.
    #  (def that people can override behavior; not a general
    #  purpose method)
    #
    def getTemplateNameFromFileName(self, fileName):
        name = fileName
        suffix = name.rfind(DEFAULT_EXTENSION)
        if suffix >= 0:
            name = name[:suffix]
        return name


    ## Define an examplar template; precompiled and stored
    #  with no attributes.  Remove any previous definition.
    #
    def defineTemplate(self, name, template):
        if name is not None and '.' in name:
            raise ValueError("cannot have '.' in template names")
        
        st = self.createStringTemplate()
        st.name = name
        st.group = self
        st.nativeGroup = self
        st.template = template
        st.errorListener = self.listener
        
        self.templates[name] = st
        return st

    def defineRegionTemplate(self, enclosingTemplate, regionName, template, type):
        """Track all references to regions <@foo>...<@end> or <@foo()>."""
        
        if isinstance(enclosingTemplate, StringTemplate):
            enclosingTemplateName = self.getMangledRegionName(
                enclosingTemplate.getOutermostName(),
                regionName
                )
            enclosingTemplate.getOutermostEnclosingInstance().addRegionName(regionName)

        else:
            enclosingTemplateName = self.getMangledRegionName(enclosingTemplate, regionName)

        regionST = self.defineTemplate(enclosingTemplateName, template)
        regionST.setIsRegion(True)
        regionST.regionDefType = type
        return regionST

    def defineImplicitRegionTemplate(self, enclosingTemplate, name):
        """
        Track all references to regions <@foo()>.  We automatically
        define as

        @enclosingtemplate.foo() ::= ""

        You cannot set these manually in the same group; you have to subgroup
        to override.
        """
        return self.defineRegionTemplate(
            enclosingTemplate,
            name,
            "",
            REGION_IMPLICIT
            )

    def getMangledRegionName(self, enclosingTemplateName, name):
        """
        The 'foo' of t() ::= '<@foo()>' is mangled to 'region__t__foo'
        """
        return "region__"+enclosingTemplateName+"__"+name

    
    def getUnMangledTemplateName(self, mangledName):
        """
        Return "t" from "region__t__foo"
        """
        return mangledName[len("region__"):mangledName.rindex("__")]


    ## Make name and alias for target.  Replace any previous def of name#
    def defineTemplateAlias(self, name, target):
        targetST = self.getTemplateDefinition(target)
        if not targetST:
            self.error('cannot alias ' + name + ' to undefined template: ' +
                       target)
            return None
        self.templates[name] = targetST
        return targetST

    def isDefinedInThisGroup(self, name):
        st = self.templates.get(name, None)
        if st is not None:
            if st.isRegion():
                # don't allow redef of @t.r() ::= "..." or <@r>...<@end>
                if st.regionDefType == REGION_IMPLICIT:
                    return False
                
            return True
        
        return False

    ## Get the ST for 'name' in this group only
    #
    def getTemplateDefinition(self, name):
        if self.templates.has_key(name):
            return self.templates[name]

    ## Is there *any* definition for template 'name' in this template
    #  or above it in the group hierarchy?
    #
    def isDefined(self, name):
        try:
            return self.lookupTemplate(name) is not None
        except ValueError:
            return False

    def parseGroup(self, r):
        try:
            lexer = GroupLexer.Lexer(r)
            parser = GroupParser.Parser(lexer)
            parser.group(self)
            # sys.stderr.write("read group\n" + str(self))
        except "foo", e: # FIXME: Exception, e:
            name = "<unknown>"
            if self.name:
                name = self.name
            self.error('problem parsing group ' + name + ': ' + str(e), e)

    def verifyInterfaceImplementations(self):
        """verify that this group satisfies its interfaces"""

        for interface in self.interfaces:
            missing = interface.getMissingTemplates(self)
            mismatched = interface.getMismatchedTemplates(self)
            
            if missing:
                self.error(
                    "group " + self.name +
                    " does not satisfy interface " + interface.name +
                    ": missing templates [" + ', '.join(["'%s'" % m for m in missing]) + ']'
                    )

            if mismatched:
                self.error(
                    "group " + self.name +
                    " does not satisfy interface " + interface.name +
                    ": mismatched arguments on these templates [" + ', '.join(["'%s'" % m for m in mismatched]) + ']'
                    )


    @deprecated
    def getRefreshInterval(self):
        return self.refreshInterval

    ## How often to refresh all templates from disk.  This is a crude
    #  mechanism at the moment--just tosses everything out at this
    #  frequency.  Set interval to 0 to refresh constantly (no caching).
    #  Set interval to a huge number like MAX_INT to have no refreshing
    #  at all (DEFAULT); it will cache stuff.
    @deprecated
    def setRefreshInterval(self, refreshInterval):
        self.refreshInterval = refreshInterval


    def setErrorListener(self, listener):
        self.listener = listener

    def getErrorListener(self):
        return self.listener

    errorListener = property(getErrorListener, setErrorListener)
    getErrorListener = deprecated(getErrorListener)
    setErrorListener = deprecated(setErrorListener)

    
    ## Specify a StringTemplateWriter implementing class to use for
    #  filtering output
    def setStringTemplateWriter(self, c):
        self.userSpecifiedWriter = c

    ## return an instance of a StringTemplateWriter that spits output to w.
    #  If a writer is specified, use it instead of the default.
    def getStringTemplateWriter(self, w):
        stw = None
        if self.userSpecifiedWriter:
            try:
                stw = self.userSpecifiedWriter(w)
            except RuntimeError, e: #FIXME Exception, e:
                self.error('problems getting StringTemplateWriter', e)

        if not stw:
            stw = AutoIndentWriter(w)
        return stw

    ## Specify a complete map of what object classes should map to which
    #  renderer objects for every template in this group (that doesn't
    #  override it per template).
    @deprecated
    def setAttributeRenderers(self, renderers):
        self.attributeRenderers = renderers

    ## Register a renderer for all objects of a particular type for all
    #  templates in this group.
    #
    def registerRenderer(self, attributeClassType, renderer):
        if not self.attributeRenderers:
            self.attributeRenderers = {}
        self.attributeRenderers[attributeClassType] = renderer

    ## What renderer is registered for this attributeClassType for
    #  this group?  If not found, as superGroup if it has one.
    #
    def getAttributeRenderer(self, attributeClassType):
        if not self.attributeRenderers:
            if not self.superGroup:
                return None # no renderers and no parent?  Stop.
            # no renderers; consult super group
            return self.superGroup.getAttributeRenderer(attributeClassType)

        if self.attributeRenderers.has_key(attributeClassType):
            return self.attributeRenderers[attributeClassType]

        elif self.superGroup is not None:
            # no renderer registered for this class, check super group
            return self.superGroup.getAttributeRenderer(attributeClassType)
        
        return None

    def getMap(self, name):
        if not self.maps:
            if not self.superGroup:
                return None
            return self.superGroup.getMap(name)
        m = None
        if self.maps.has_key(name):
            m = self.maps[name]
        if (not m) and self.superGroup:
            m = self.superGroup.getMap(name)
        return m

    def defineMap(self, name, mapping):
        """
        Define a map for this group; not thread safe...do not keep adding
 	these while you reference them.
        """
        
        self.maps[name] = mapping

    @classmethod
    def registerGroupLoader(cls, loader):
        cls.groupLoader = loader

    @classmethod
    def registerDefaultLexer(cls, lexerClass):
        cls.defaultTemplateLexerClass = lexerClass

    @classmethod
    def loadGroup(cls, name, superGroup=None, lexer=None):
        if cls.groupLoader is not None:
            return cls.groupLoader.loadGroup(name, superGroup, lexer)

        return None
    
    @classmethod
    def loadInterface(cls, name):
        if cls.groupLoader is not None:
            return cls.groupLoader.loadInterface(name)

        return None
    
    def error(self, msg, e = None):
        if self.listener:
            self.listener.error(msg, e)
        else:
            sys.stderr.write('StringTemplate: ' + msg + ': ' + e + '\n')
            traceback.print_exc()
            
    def getTemplateNames(self):
        return self.templates.keys()


    def emitDebugStartStopStrings(self, emit):
        """
        Indicate whether ST should emit <templatename>...</templatename>
        strings for debugging around output for templates from this group.
        """
        
        self.debugTemplateOutput = emit


    def doNotEmitDebugStringsForTemplate(self, templateName):
        if self.noDebugStartStopStrings is None:
            self.noDebugStartStopStrings = set()

        self.noDebugStartStopStrings.add(templateName)


    def emitTemplateStartDebugString(self, st, out):
        if (self.noDebugStartStopStrings is None or
            st.name not in self.noDebugStartStopStrings ):
            out.write("<"+st.name+">")


    def emitTemplateStopDebugString(self, st, out):
        if (self.noDebugStartStopStrings is None or
            st.name not in self.noDebugStartStopStrings ):
            out.write("</"+st.name+">")


    def toString(self, showTemplatePatterns=True):
        buf = StringIO()
        buf.write('group ' + str(self.name) + ';\n')
        sortedNames = self.templates.keys()
        sortedNames.sort()
        for tname in sortedNames:
            st = self.templates[tname]
            if st != StringTemplateGroup.NOT_FOUND_ST:
                args = st.formalArguments.keys()
                args.sort()
                buf.write(str(tname) + '(' + ",".join(args) + ')')
                if showTemplatePatterns:
                    buf.write(' ::= <<' + str(st.template) + '>>\n')
                else:
                    buf.write('\n')

        retval = buf.getvalue()
        buf.close()
        return retval

    __str__ = toString




# [The "BSD licence"]
# Copyright (c) 2003-2006 Terence Parr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from StringIO import StringIO

from stringtemplate3.language import (
    InterfaceLexer, InterfaceParser
    )

from stringtemplate3.utils import deprecated
from stringtemplate3.errors import DEFAULT_ERROR_LISTENER


class TemplateDefinition(object):
    """All the info we need to track for a template defined in an interface"""

    def __init__(self, name, formalArgs, optional):
        self.name = name
        self.formalArgs = formalArgs
        self.optional = optional


class StringTemplateGroupInterface(object):
    """
    A group interface is like a group without the template implementations;
    there are just template names/argument-lists like this:

    interface foo;
    class(name,fields);
    method(name,args,body);
    """

    def __init__(self, file, errors=None, superInterface=None):
        """Create an interface from the input stream"""

	## What is the group name
        self.name = None

        ## Maps template name to TemplateDefinition object
        self.templates = {}

	## Are we derived from another group?  Templates not found in this
        # group will be searched for in the superGroup recursively.
        self.superInterface = superInterface

	## Where to report errors.  All string templates in this group
        #  use this error handler by default.
        if errors is not None:
            self.listener = errors
        else:
            self.listener = DEFAULT_ERROR_LISTENER

        self.parseInterface(file)


    @deprecated
    def getSuperInterface(self):
        return self.superInterface

    @deprecated
    def setSuperInterface(self, superInterface):
        self.superInterface = superInterface


    def parseInterface(self, r):
        try:
            lexer = InterfaceLexer.Lexer(r)
            parser = InterfaceParser.Parser(lexer)
            parser.groupInterface(self)

        except RuntimeError, exc: #FIXME:  Exception
            name = self.name or "<unknown>"
            self.error("problem parsing group "+name+": "+str(exc), exc)


    def defineTemplate(self, name, formalArgs, optional):
        d = TemplateDefinition(name, formalArgs, optional)
        self.templates[d.name] = d


    def getMissingTemplates(self, group):
        """
        Return a list of all template names missing from group that are defined
        in this interface.  Return null if all is well.
        """

        missing = []

        templates = self.templates.items()
        templates.sort()
        for name, d in templates:
            if not d.optional and not group.isDefined(d.name):
                missing.append(d.name)

        return missing or None


    def getMismatchedTemplates(self, group):
        """
        Return a list of all template sigs that are present in the group, but
        that have wrong formal argument lists.  Return null if all is well.
        """

        mismatched = []
        
        templates = self.templates.items()
        templates.sort()
        for name, d in templates:
            if group.isDefined(d.name):
                defST = group.getTemplateDefinition(d.name)
                formalArgs = defST.formalArguments
                ack = False
                if ( (d.formalArgs is not None and formalArgs is None) or
                     (d.formalArgs is None and formalArgs is not None) or
                     (len(d.formalArgs) != len(formalArgs)) ):
                    ack = True

                if not ack:
                    for argName in formalArgs.keys():
                        if d.formalArgs.get(argName, None) == None:
                            ack = True
                            break

                if ack:
                    mismatched.append(self.getTemplateSignature(d))

        return mismatched or None


    @deprecated
    def getName(self):
        return self.name

    @deprecated
    def setName(self, name):
        self.name = name


    def error(self, msg, exc=None):
        if self.listener is not None:
            self.listener.error(msg, exc)


    def toString(self):
        buf = StringIO()
        buf.write("interface ")
        buf.write(self.name)
        buf.write(";\n")
        templates = self.templates.items()
        templates.sort()
        for name, d in templates:
            buf.write( self.getTemplateSignature(d) )
            buf.write(";\n")

        return buf.getvalue()
    
    __str__ = toString

    
    def getTemplateSignature(self, d):
        buf = StringIO()
        if d.optional:
            buf.write("optional ")

        buf.write(d.name)
        if d.formalArgs is not None:
            buf.write('(')
            args = d.formalArgs.keys()
            args.sort()
            buf.write(", ".join(args))
            buf.write(')')

        else:
            buf.write("()")

        return buf.getvalue()





# the following represent bit positions emulating a cardinality bitset.
OPTIONAL = 1      # a?
REQUIRED = 2      # a
ZERO_OR_MORE = 4  # a*
ONE_OR_MORE = 8   # a+

suffixes = [
    None,
    "?",
    "",
    None,
    "*",
    None,
    None,
    None,
    "+"
    ]

## When template arguments are not available such as when the user
#  uses "new StringTemplate(...)", then the list of formal arguments
#  must be distinguished from the case where a template can specify
#  args and there just aren't any such as the t() template above.
UNKNOWN_ARGS = { '<unknown>': -1 }


def getCardinalityName(cardinality):
    return {
        OPTIONAL: 'optional',
        REQUIRED: 'exactly one',
        ZERO_OR_MORE: 'zero-or-more',
        ONE_OR_MORE: 'one-or-more'
        }.get(cardinality, 'unknown')
    

class FormalArgument(object):
    """
    Represents the name of a formal argument
    defined in a template:

    group test;
    test(a,b) : "$a$ $b$"
    t() : "blort"

    Each template has a set of these formal arguments or uses
    a placeholder object: UNKNOWN (indicating that no arguments
    were specified such as when a template is loaded from a file.st).

    Note: originally, I tracked cardinality as well as the name of an
    attribute.  I'm leaving the code here as I suspect something may come
    of it later.  Currently, though, cardinality is not used.
    """
    
    def __init__(self, name, defaultValueST = None):
        self.name = name
        # self.cardinality = REQUIRED
        ## If they specified name="value", store the template here
        #
        self.defaultValueST = defaultValueST


    def __eq__(self, other):
        if not isinstance(other, FormalArgument):
            return False
        
        if self.name != other.name:
            return False

        # only check if there is a default value; that's all
        if ( (self.defaultValueST is not None
              and other.defaultValueST is None) or
             (self.defaultValueST is None
              and other.defaultValueST is not None) ):
            return False

        return True


    def __neq__(self, other):
        return not self.__eq__(other)

    
    def __str__(self):
        if self.defaultValueST:
            return self.name + '=' + str(self.defaultValueST)
        return self.name

    __repr__ = __str__




### $ANTLR 2.7.7 (2006-11-01): "eval.g" -> "ActionEvaluator.py"$
### import antlr and other modules ..
import sys
import antlr

version = sys.version.split()[0]
if version < '2.2.1':
    False = 0
if version < '2.3':
    True = not False
### header action >>> 
from stringtemplate3.language.CatIterator import CatList
from stringtemplate3.language.StringTemplateAST import StringTemplateAST

from StringIO import StringIO

class NameValuePair(object):

   def __init__(self):
       self.name = None
       self.value = None
### header action <<< 

### import antlr.Token 
from antlr import Token
### >>>The Known Token Types <<<
SKIP                = antlr.SKIP
INVALID_TYPE        = antlr.INVALID_TYPE
EOF_TYPE            = antlr.EOF_TYPE
EOF                 = antlr.EOF
NULL_TREE_LOOKAHEAD = antlr.NULL_TREE_LOOKAHEAD
MIN_USER_TYPE       = antlr.MIN_USER_TYPE
APPLY = 4
MULTI_APPLY = 5
ARGS = 6
INCLUDE = 7
CONDITIONAL = 8
VALUE = 9
TEMPLATE = 10
FUNCTION = 11
SINGLEVALUEARG = 12
LIST = 13
NOTHING = 14
SEMI = 15
LPAREN = 16
RPAREN = 17
LITERAL_elseif = 18
COMMA = 19
ID = 20
ASSIGN = 21
COLON = 22
NOT = 23
PLUS = 24
DOT = 25
LITERAL_first = 26
LITERAL_rest = 27
LITERAL_last = 28
LITERAL_length = 29
LITERAL_strip = 30
LITERAL_trunc = 31
LITERAL_super = 32
ANONYMOUS_TEMPLATE = 33
STRING = 34
INT = 35
LBRACK = 36
RBRACK = 37
DOTDOTDOT = 38
TEMPLATE_ARGS = 39
NESTED_ANONYMOUS_TEMPLATE = 40
ESC_CHAR = 41
WS = 42
WS_CHAR = 43

### user code>>>

### user code<<<

class Walker(antlr.TreeParser):
    
    # ctor ..
    def __init__(self, *args, **kwargs):
        antlr.TreeParser.__init__(self, *args, **kwargs)
        self.tokenNames = _tokenNames
        ### __init__ header action >>> 
        self.this = None
        self.out = None
        self.chunk = None
        ### __init__ header action <<< 
    
    ### user action >>>
    def initialize(self, this, chunk, out):
       self.this = this
       self.chunk = chunk
       self.out = out
    
    def reportError(self, e):
       self.this.error("eval tree parse error", e)
    ### user action <<<
    def action(self, _t):    
        numCharsWritten = 0
        
        action_AST_in = None
        if _t != antlr.ASTNULL:
            action_AST_in = _t
        e = None
        try:      ## for error handling
            pass
            e=self.expr(_t)
            _t = self._retTree
            numCharsWritten = self.chunk.writeAttribute(self.this, e, self.out)
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return numCharsWritten
    
    def expr(self, _t):    
        value = None
        
        expr_AST_in = None
        if _t != antlr.ASTNULL:
            expr_AST_in = _t
        a = None
        b = None
        e = None
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [PLUS]:
                pass
                _t3 = _t
                tmp1_AST_in = _t
                self.match(_t,PLUS)
                _t = _t.getFirstChild()
                a=self.expr(_t)
                _t = self._retTree
                b=self.expr(_t)
                _t = self._retTree
                value = self.chunk.add(a,b)
                _t = _t3
                _t = _t.getNextSibling()
            elif la1 and la1 in [APPLY,MULTI_APPLY]:
                pass
                value=self.templateApplication(_t)
                _t = self._retTree
            elif la1 and la1 in [ID,DOT,ANONYMOUS_TEMPLATE,STRING,INT]:
                pass
                value=self.attribute(_t)
                _t = self._retTree
            elif la1 and la1 in [INCLUDE]:
                pass
                value=self.templateInclude(_t)
                _t = self._retTree
            elif la1 and la1 in [FUNCTION]:
                pass
                value=self.function(_t)
                _t = self._retTree
            elif la1 and la1 in [LIST,NOTHING]:
                pass
                value=self.list(_t)
                _t = self._retTree
            elif la1 and la1 in [VALUE]:
                pass
                _t4 = _t
                tmp2_AST_in = _t
                self.match(_t,VALUE)
                _t = _t.getFirstChild()
                e=self.expr(_t)
                _t = self._retTree
                _t = _t4
                _t = _t.getNextSibling()
                buf = StringIO()
                sw = self.this.group.getStringTemplateWriter(buf)
                n = self.chunk.writeAttribute(self.this, e, sw)
                if n > 0:
                   value = buf.getvalue()
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    

    ###/** Apply template(s) to an attribute; can be applied to another apply
    ### *  result.
    ### */
    def templateApplication(self, _t):    
        value = None
        
        templateApplication_AST_in = None
        if _t != antlr.ASTNULL:
            templateApplication_AST_in = _t
        anon = None
        a = None
        templatesToApply = []
        attributes = []
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [APPLY]:
                pass
                _t14 = _t
                tmp3_AST_in = _t
                self.match(_t,APPLY)
                _t = _t.getFirstChild()
                a=self.expr(_t)
                _t = self._retTree
                _cnt16= 0
                while True:
                    if not _t:
                        _t = antlr.ASTNULL
                    if (_t.getType()==TEMPLATE):
                        pass
                        self.template(_t, templatesToApply)
                        _t = self._retTree
                    else:
                        break
                    
                    _cnt16 += 1
                if _cnt16 < 1:
                    raise antlr.NoViableAltException(_t)
                value = self.chunk.applyListOfAlternatingTemplates(self.this, \
                   a, templatesToApply)
                _t = _t14
                _t = _t.getNextSibling()
            elif la1 and la1 in [MULTI_APPLY]:
                pass
                _t17 = _t
                tmp4_AST_in = _t
                self.match(_t,MULTI_APPLY)
                _t = _t.getFirstChild()
                _cnt19= 0
                while True:
                    if not _t:
                        _t = antlr.ASTNULL
                    if (_tokenSet_0.member(_t.getType())):
                        pass
                        a=self.expr(_t)
                        _t = self._retTree
                        attributes.append(a)
                    else:
                        break
                    
                    _cnt19 += 1
                if _cnt19 < 1:
                    raise antlr.NoViableAltException(_t)
                tmp5_AST_in = _t
                self.match(_t,COLON)
                _t = _t.getNextSibling()
                anon = _t
                self.match(_t,ANONYMOUS_TEMPLATE)
                _t = _t.getNextSibling()
                anonymous = anon.getStringTemplate()
                templatesToApply.append(anonymous)
                value = self.chunk.applyTemplateToListOfAttributes(self.this, \
                   attributes, anon.getStringTemplate())
                _t = _t17
                _t = _t.getNextSibling()
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def attribute(self, _t):    
        value = None
        
        attribute_AST_in = None
        if _t != antlr.ASTNULL:
            attribute_AST_in = _t
        prop = None
        i3 = None
        i = None
        s = None
        at = None
        obj = None
        propName = None
        e = None
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [DOT]:
                pass
                _t33 = _t
                tmp6_AST_in = _t
                self.match(_t,DOT)
                _t = _t.getFirstChild()
                obj=self.expr(_t)
                _t = self._retTree
                if not _t:
                    _t = antlr.ASTNULL
                la1 = _t.getType()
                if False:
                    pass
                elif la1 and la1 in [ID]:
                    pass
                    prop = _t
                    self.match(_t,ID)
                    _t = _t.getNextSibling()
                    propName = prop.getText()
                elif la1 and la1 in [VALUE]:
                    pass
                    _t35 = _t
                    tmp7_AST_in = _t
                    self.match(_t,VALUE)
                    _t = _t.getFirstChild()
                    e=self.expr(_t)
                    _t = self._retTree
                    _t = _t35
                    _t = _t.getNextSibling()
                    if e: propName = str(e)
                else:
                        raise antlr.NoViableAltException(_t)
                    
                _t = _t33
                _t = _t.getNextSibling()
                value = self.chunk.getObjectProperty(self.this, obj, propName)
            elif la1 and la1 in [ID]:
                pass
                i3 = _t
                self.match(_t,ID)
                _t = _t.getNextSibling()
                value = self.this.getAttribute(i3.getText())
            elif la1 and la1 in [INT]:
                pass
                i = _t
                self.match(_t,INT)
                _t = _t.getNextSibling()
                value = int(i.getText())
            elif la1 and la1 in [STRING]:
                pass
                s = _t
                self.match(_t,STRING)
                _t = _t.getNextSibling()
                value = s.getText()
            elif la1 and la1 in [ANONYMOUS_TEMPLATE]:
                pass
                at = _t
                self.match(_t,ANONYMOUS_TEMPLATE)
                _t = _t.getNextSibling()
                value = at.getText();
                if at.getText():
                   from stringtemplate3.templates import StringTemplate
                   valueST = StringTemplate(
                       group=self.this.group,
                       template=at.getText()
                       )
                   valueST.enclosingInstance = self.this
                   valueST.name = "<anonymous template argument>"
                   value = valueST
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def templateInclude(self, _t):    
        value = None
        
        templateInclude_AST_in = None
        if _t != antlr.ASTNULL:
            templateInclude_AST_in = _t
        id = None
        a1 = None
        a2 = None
        args = None
        name = ""
        n = None
        try:      ## for error handling
            pass
            _t10 = _t
            tmp8_AST_in = _t
            self.match(_t,INCLUDE)
            _t = _t.getFirstChild()
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [ID]:
                pass
                id = _t
                self.match(_t,ID)
                _t = _t.getNextSibling()
                a1 = _t
                if not _t:
                    raise antlr.MismatchedTokenException()
                _t = _t.getNextSibling()
                name=id.getText(); args=a1
            elif la1 and la1 in [VALUE]:
                pass
                _t12 = _t
                tmp9_AST_in = _t
                self.match(_t,VALUE)
                _t = _t.getFirstChild()
                n=self.expr(_t)
                _t = self._retTree
                a2 = _t
                if not _t:
                    raise antlr.MismatchedTokenException()
                _t = _t.getNextSibling()
                _t = _t12
                _t = _t.getNextSibling()
                if n: name = str(n); args=a2
            else:
                    raise antlr.NoViableAltException(_t)
                
            _t = _t10
            _t = _t.getNextSibling()
            if name:
               value = self.chunk.getTemplateInclude(self.this, name, args)
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def function(self, _t):    
        value = None
        
        function_AST_in = None
        if _t != antlr.ASTNULL:
            function_AST_in = _t
        a = None
        try:      ## for error handling
            pass
            _t21 = _t
            tmp10_AST_in = _t
            self.match(_t,FUNCTION)
            _t = _t.getFirstChild()
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [LITERAL_first]:
                pass
                tmp11_AST_in = _t
                self.match(_t,LITERAL_first)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.first(a)
            elif la1 and la1 in [LITERAL_rest]:
                pass
                tmp12_AST_in = _t
                self.match(_t,LITERAL_rest)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.rest(a)
            elif la1 and la1 in [LITERAL_last]:
                pass
                tmp13_AST_in = _t
                self.match(_t,LITERAL_last)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.last(a)
            elif la1 and la1 in [LITERAL_length]:
                pass
                tmp14_AST_in = _t
                self.match(_t,LITERAL_length)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.length(a)
            elif la1 and la1 in [LITERAL_strip]:
                pass
                tmp15_AST_in = _t
                self.match(_t,LITERAL_strip)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.strip(a)
            elif la1 and la1 in [LITERAL_trunc]:
                pass
                tmp16_AST_in = _t
                self.match(_t,LITERAL_trunc)
                _t = _t.getNextSibling()
                a=self.singleFunctionArg(_t)
                _t = self._retTree
                value = self.chunk.trunc(a)
            else:
                    raise antlr.NoViableAltException(_t)
                
            _t = _t21
            _t = _t.getNextSibling()
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    

    ###/** create a new list of expressions as a new multi-value attribute */
    def list(self, _t):    
        value=None
        
        list_AST_in = None
        if _t != antlr.ASTNULL:
            list_AST_in = _t
        e = None
        elements = []
        value = CatList(elements)
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [LIST]:
                pass
                _t6 = _t
                tmp17_AST_in = _t
                self.match(_t,LIST)
                _t = _t.getFirstChild()
                _cnt8= 0
                while True:
                    if not _t:
                        _t = antlr.ASTNULL
                    if (_tokenSet_0.member(_t.getType())):
                        pass
                        e=self.expr(_t)
                        _t = self._retTree
                        if e:
                           from stringtemplate3.language.ASTExpr import convertAnythingToList
                           e = convertAnythingToList(e)
                           elements.append(e)
                    else:
                        break
                    
                    _cnt8 += 1
                if _cnt8 < 1:
                    raise antlr.NoViableAltException(_t)
                _t = _t6
                _t = _t.getNextSibling()
            elif la1 and la1 in [NOTHING]:
                pass
                tmp18_AST_in = _t
                self.match(_t,NOTHING)
                _t = _t.getNextSibling()
                nullSingleton = [None]
                element.append(nullSingleton)
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def template(self, _t,
        templatesToApply
    ):    
        
        template_AST_in = None
        if _t != antlr.ASTNULL:
            template_AST_in = _t
        t = None
        args = None
        anon = None
        args2 = None
        argumentContext = {}
        n = None
        try:      ## for error handling
            pass
            _t26 = _t
            tmp19_AST_in = _t
            self.match(_t,TEMPLATE)
            _t = _t.getFirstChild()
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [ID]:
                pass
                t = _t
                self.match(_t,ID)
                _t = _t.getNextSibling()
                args = _t
                if not _t:
                    raise antlr.MismatchedTokenException()
                _t = _t.getNextSibling()
                templateName = t.getText()
                group = self.this.group
                embedded = group.getEmbeddedInstanceOf(
                      templateName,
                      self.this
                  )
                if embedded:
                   embedded.argumentsAST = args
                   templatesToApply.append(embedded)
            elif la1 and la1 in [ANONYMOUS_TEMPLATE]:
                pass
                anon = _t
                self.match(_t,ANONYMOUS_TEMPLATE)
                _t = _t.getNextSibling()
                anonymous = anon.getStringTemplate()
                # to properly see overridden templates, always set
                # anonymous' group to be self's group
                anonymous.group = self.this.group
                templatesToApply.append(anonymous)
            elif la1 and la1 in [VALUE]:
                pass
                _t28 = _t
                tmp20_AST_in = _t
                self.match(_t,VALUE)
                _t = _t.getFirstChild()
                n=self.expr(_t)
                _t = self._retTree
                args2 = _t
                if not _t:
                    raise antlr.MismatchedTokenException()
                _t = _t.getNextSibling()
                embedded = None
                if n:
                   templateName = str(n)
                   group = self.this.group
                   embedded = group.getEmbeddedInstanceOf(
                       templateName,
                       self.this
                       )
                   if embedded:
                       embedded.argumentsAST = args2
                       templatesToApply.append(embedded)
                _t = _t28
                _t = _t.getNextSibling()
            else:
                    raise antlr.NoViableAltException(_t)
                
            _t = _t26
            _t = _t.getNextSibling()
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
    
    def singleFunctionArg(self, _t):    
        value = None
        
        singleFunctionArg_AST_in = None
        if _t != antlr.ASTNULL:
            singleFunctionArg_AST_in = _t
        try:      ## for error handling
            pass
            _t24 = _t
            tmp21_AST_in = _t
            self.match(_t,SINGLEVALUEARG)
            _t = _t.getFirstChild()
            value=self.expr(_t)
            _t = self._retTree
            _t = _t24
            _t = _t.getNextSibling()
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def ifCondition(self, _t):    
        value = False
        
        ifCondition_AST_in = None
        if _t != antlr.ASTNULL:
            ifCondition_AST_in = _t
        a = None
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [APPLY,MULTI_APPLY,INCLUDE,VALUE,FUNCTION,LIST,NOTHING,ID,PLUS,DOT,ANONYMOUS_TEMPLATE,STRING,INT]:
                pass
                a=self.ifAtom(_t)
                _t = self._retTree
                value = self.chunk.testAttributeTrue(a)
            elif la1 and la1 in [NOT]:
                pass
                _t30 = _t
                tmp22_AST_in = _t
                self.match(_t,NOT)
                _t = _t.getFirstChild()
                a=self.ifAtom(_t)
                _t = self._retTree
                _t = _t30
                _t = _t.getNextSibling()
                value = not self.chunk.testAttributeTrue(a)
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    
    def ifAtom(self, _t):    
        value = None
        
        ifAtom_AST_in = None
        if _t != antlr.ASTNULL:
            ifAtom_AST_in = _t
        try:      ## for error handling
            pass
            value=self.expr(_t)
            _t = self._retTree
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return value
    

    ###/** self is assumed to be the enclosing context as foo(x=y) must find y in
    ### *  the template that encloses the ref to foo(x=y).  We must pass in
    ### *  the embedded template (the one invoked) so we can check formal args
    ### *  in rawSetArgumentAttribute.
    ### */
    def argList(self, _t,
        embedded, initialContext
    ):    
        argumentContext = None
        
        argList_AST_in = None
        if _t != antlr.ASTNULL:
            argList_AST_in = _t
        argumentContext = initialContext
        if not argumentContext:
           argumentContext = {}
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [ARGS]:
                pass
                _t37 = _t
                tmp23_AST_in = _t
                self.match(_t,ARGS)
                _t = _t.getFirstChild()
                while True:
                    if not _t:
                        _t = antlr.ASTNULL
                    if (_t.getType()==ASSIGN or _t.getType()==DOTDOTDOT):
                        pass
                        self.argumentAssignment(_t, embedded, argumentContext)
                        _t = self._retTree
                    else:
                        break
                    
                _t = _t37
                _t = _t.getNextSibling()
            elif la1 and la1 in [SINGLEVALUEARG]:
                pass
                self.singleTemplateArg(_t, embedded, argumentContext)
                _t = self._retTree
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
        return argumentContext
    
    def argumentAssignment(self, _t,
        embedded, argumentContext
    ):    
        
        argumentAssignment_AST_in = None
        if _t != antlr.ASTNULL:
            argumentAssignment_AST_in = _t
        arg = None
        e = None
        try:      ## for error handling
            if not _t:
                _t = antlr.ASTNULL
            la1 = _t.getType()
            if False:
                pass
            elif la1 and la1 in [ASSIGN]:
                pass
                _t43 = _t
                tmp24_AST_in = _t
                self.match(_t,ASSIGN)
                _t = _t.getFirstChild()
                arg = _t
                self.match(_t,ID)
                _t = _t.getNextSibling()
                e=self.expr(_t)
                _t = self._retTree
                _t = _t43
                _t = _t.getNextSibling()
                if e:
                   self.this.rawSetArgumentAttribute(embedded, argumentContext, \
                       arg.getText(), e)
            elif la1 and la1 in [DOTDOTDOT]:
                pass
                tmp25_AST_in = _t
                self.match(_t,DOTDOTDOT)
                _t = _t.getNextSibling()
                embedded.passThroughAttributes = True
            else:
                    raise antlr.NoViableAltException(_t)
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
    
    def singleTemplateArg(self, _t,
        embedded, argumentContext
    ):    
        
        singleTemplateArg_AST_in = None
        if _t != antlr.ASTNULL:
            singleTemplateArg_AST_in = _t
        e = None
        try:      ## for error handling
            pass
            _t41 = _t
            tmp26_AST_in = _t
            self.match(_t,SINGLEVALUEARG)
            _t = _t.getFirstChild()
            e=self.expr(_t)
            _t = self._retTree
            _t = _t41
            _t = _t.getNextSibling()
            if e:
               soleArgName = None
               # find the sole defined formal argument for embedded
               error = False
               formalArgs = embedded.formalArguments
               if formalArgs:
                   argNames = formalArgs.keys()
                   if len(argNames) == 1:
                       soleArgName = argNames[0]
                       #sys.stderr.write("sole formal arg of " +
                       #                 embedded.getName() + " is " +
                       #                 soleArgName)
                   else:
                       error = True
            else:
               error = True
            if error:
               self.this.error("template " + embedded.name +
                               " must have exactly one formal arg in" +
                               " template context " +
                               self.this.enclosingInstanceStackString)
            else:
               self.this.rawSetArgumentAttribute(embedded, \
                   argumentContext, soleArgName, e)
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            if _t:
                _t = _t.getNextSibling()
        
        self._retTree = _t
    

_tokenNames = [
    "<0>", 
    "EOF", 
    "<2>", 
    "NULL_TREE_LOOKAHEAD", 
    "APPLY", 
    "MULTI_APPLY", 
    "ARGS", 
    "INCLUDE", 
    "\"if\"", 
    "VALUE", 
    "TEMPLATE", 
    "FUNCTION", 
    "SINGLEVALUEARG", 
    "LIST", 
    "NOTHING", 
    "SEMI", 
    "LPAREN", 
    "RPAREN", 
    "\"elseif\"", 
    "COMMA", 
    "ID", 
    "ASSIGN", 
    "COLON", 
    "NOT", 
    "PLUS", 
    "DOT", 
    "\"first\"", 
    "\"rest\"", 
    "\"last\"", 
    "\"length\"", 
    "\"strip\"", 
    "\"trunc\"", 
    "\"super\"", 
    "ANONYMOUS_TEMPLATE", 
    "STRING", 
    "INT", 
    "LBRACK", 
    "RBRACK", 
    "DOTDOTDOT", 
    "TEMPLATE_ARGS", 
    "NESTED_ANONYMOUS_TEMPLATE", 
    "ESC_CHAR", 
    "WS", 
    "WS_CHAR"
]
    

### generate bit set
def mk_tokenSet_0(): 
    ### var1
    data = [ 60180949680L, 0L]
    return data
_tokenSet_0 = antlr.BitSet(mk_tokenSet_0())




import antlr

class StringTemplateToken(antlr.CommonToken):

    def __init__(self, type = 0, text = '', args = None):
        super(StringTemplateToken, self).__init__(type=type, text=text)
        ## Track any args for anonymous templates like
        #  <tokens,rules:{t,r | <t> then <r>}>
        #  The lexer in action.g returns a single token ANONYMOUS_TEMPLATE
        #  and so I need to have it parse args in the lexer and make them
        #  available for when I build the anonymous template.

        if args is None:
            args = []
        self.args = args

    def __str__(self):
        return super(StringTemplateToken, self).__str__() + \
               '; args=' + str(self.args)

    __repr__ = __str__




from stringtemplate3.language.StringRef import StringRef

class NewlineRef(StringRef):
    """
    Represents a newline.  Separated so I can do smart things like not
    spitting out newlines when the only thing on a line is an attr expr.
    """
    
    def __init__(self, enclosingTemplate, str):
        super(NewlineRef, self).__init__(enclosingTemplate, str)



### $ANTLR 2.7.7 (2006-11-01): "angle.bracket.template.g" -> "AngleBracketTemplateLexer.py"$
### import antlr and other modules ..
import sys
import antlr

version = sys.version.split()[0]
if version < '2.2.1':
    False = 0
if version < '2.3':
    True = not False
### header action >>> 
import stringtemplate3
import stringtemplate3.language.TemplateParser
from stringtemplate3.language.ChunkToken import ChunkToken
### header action <<< 
### preamble action >>> 

### preamble action <<< 
### >>>The Literals<<<
literals = {}


### import antlr.Token 
from antlr import Token
### >>>The Known Token Types <<<
SKIP                = antlr.SKIP
INVALID_TYPE        = antlr.INVALID_TYPE
EOF_TYPE            = antlr.EOF_TYPE
EOF                 = antlr.EOF
NULL_TREE_LOOKAHEAD = antlr.NULL_TREE_LOOKAHEAD
MIN_USER_TYPE       = antlr.MIN_USER_TYPE
LITERAL = 4
NEWLINE = 5
ACTION = 6
IF = 7
ELSEIF = 8
ELSE = 9
ENDIF = 10
REGION_REF = 11
REGION_DEF = 12
NL = 13
EXPR = 14
TEMPLATE = 15
IF_EXPR = 16
ESC_CHAR = 17
ESC = 18
HEX = 19
SUBTEMPLATE = 20
NESTED_PARENS = 21
INDENT = 22
COMMENT = 23


###/** Break up an input text stream into chunks of either plain text
### *  or template actions in "<...>".  Treat IF and ENDIF tokens
### *  specially.
### */
class Lexer(antlr.CharScanner) :
    ### user action >>>
    def reportError(self, e):
       self.this.error("<...> chunk lexer error", e)
    
    def upcomingELSE(self, i):
       return self.LA(i) == '<' and \
              self.LA(i+1) == 'e' and \
              self.LA(i+2) == 'l' and \
              self.LA(i+3) == 's' and \
              self.LA(i+4) == 'e' and \
              self.LA(i+5) == '>'
    
    def upcomingENDIF(self, i):
       return self.LA(i) == '<' and \
              self.LA(i+1) == 'e' and \
              self.LA(i+2) == 'n' and \
              self.LA(i+3) == 'd' and \
              self.LA(i+4) == 'i' and \
              self.LA(i+5) == 'f' and \
              self.LA(i+6) == '>'
    
    def upcomingAtEND(self, i):
       return self.LA(i) == '<' and \
              self.LA(i+1) == '@' and \
              self.LA(i+2) == 'e' and \
              self.LA(i+3) == 'n' and \
              self.LA(i+4) == 'd' and \
              self.LA(i+5) == '>'
    
    def upcomingNewline(self, i):
       return (self.LA(i) == '\r' and
               self.LA(i+1) == '\n') or \
              self.LA(i) == '\n'
    ### user action <<<
    def __init__(self, *argv, **kwargs) :
        antlr.CharScanner.__init__(self, *argv, **kwargs)
        self.caseSensitiveLiterals = True
        self.setCaseSensitive(True)
        self.literals = literals
        ### __init__ header action >>> 
        self.currentIndent = None
        self.this = None
        ### __init__ header action <<< 
    
    def nextToken(self):
        while True:
            try: ### try again ..
                while True:
                    _token = None
                    _ttype = INVALID_TYPE
                    self.resetText()
                    try: ## for char stream error handling
                        try: ##for lexical error handling
                            la1 = self.LA(1)
                            if False:
                                pass
                            elif la1 and la1 in u'\n\r':
                                pass
                                self.mNEWLINE(True)
                                theRetToken = self._returnToken
                            elif la1 and la1 in u'<':
                                pass
                                self.mACTION(True)
                                theRetToken = self._returnToken
                            else:
                                if ((_tokenSet_0.member(self.LA(1))) and ( self.LA(1) != '\r' and self.LA(1) != '\n' )):
                                    pass
                                    self.mLITERAL(True)
                                    theRetToken = self._returnToken
                                else:
                                    self.default(self.LA(1))
                                
                            if not self._returnToken:
                                raise antlr.TryAgain ### found SKIP token
                            ### option { testLiterals=true } 
                            self.testForLiteral(self._returnToken)
                            ### return token to caller
                            return self._returnToken
                        ### handle lexical errors ....
                        except antlr.RecognitionException, e:
                            raise antlr.TokenStreamRecognitionException(e)
                    ### handle char stream errors ...
                    except antlr.CharStreamException,cse:
                        if isinstance(cse, antlr.CharStreamIOException):
                            raise antlr.TokenStreamIOException(cse.io)
                        else:
                            raise antlr.TokenStreamException(str(cse))
            except antlr.TryAgain:
                pass
        
    def mLITERAL(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = LITERAL
        _saveIndex = 0
        ind = None
        if not  self.LA(1) != '\r' and self.LA(1) != '\n' :
            raise antlr.SemanticException(" self.LA(1) != '\\r' and self.LA(1) != '\\n' ")
        pass
        _cnt5= 0
        while True:
            loopStartIndex = self.text.length()
            col = self.getColumn()
            if (self.LA(1)==u'\\') and (self.LA(2)==u'<'):
                pass
                _saveIndex = self.text.length()
                self.match('\\')
                self.text.setLength(_saveIndex)
                self.match('<')
            elif (self.LA(1)==u'\\') and (self.LA(2)==u'>') and (True) and (True) and (True) and (True) and (True):
                pass
                _saveIndex = self.text.length()
                self.match('\\')
                self.text.setLength(_saveIndex)
                self.match('>')
            elif (self.LA(1)==u'\\') and (self.LA(2)==u'\\') and (True) and (True) and (True) and (True) and (True):
                pass
                _saveIndex = self.text.length()
                self.match('\\')
                self.text.setLength(_saveIndex)
                self.match('\\')
            elif (self.LA(1)==u'\\') and (_tokenSet_1.member(self.LA(2))) and (True) and (True) and (True) and (True) and (True):
                pass
                self.match('\\')
                self.match(_tokenSet_1)
            elif (self.LA(1)==u'\t' or self.LA(1)==u' ') and (True) and (True) and (True) and (True) and (True) and (True):
                pass
                self.mINDENT(True)
                ind = self._returnToken
                if col == 1 and self.LA(1) == '<':
                   # store indent in ASTExpr not in a literal
                   self.currentIndent = ind.getText()
                   # reset length to wack text
                   self.text.setLength(loopStartIndex)
                else:
                   self.currentIndent = None
            elif (_tokenSet_0.member(self.LA(1))) and (True) and (True) and (True) and (True) and (True) and (True):
                pass
                self.match(_tokenSet_0)
            else:
                break
            
            _cnt5 += 1
        if _cnt5 < 1:
            self.raise_NoViableAlt(self.LA(1))
        if not len(self.text.getString(_begin)): _ttype = SKIP
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mINDENT(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = INDENT
        _saveIndex = 0
        pass
        _cnt8= 0
        while True:
            if (self.LA(1)==u' ') and (True) and (True) and (True) and (True) and (True) and (True):
                pass
                self.match(' ')
            elif (self.LA(1)==u'\t') and (True) and (True) and (True) and (True) and (True) and (True):
                pass
                self.match('\t')
            else:
                break
            
            _cnt8 += 1
        if _cnt8 < 1:
            self.raise_NoViableAlt(self.LA(1))
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mNEWLINE(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = NEWLINE
        _saveIndex = 0
        pass
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in u'\r':
            pass
            self.match('\r')
        elif la1 and la1 in u'\n':
            pass
        else:
                self.raise_NoViableAlt(self.LA(1))
            
        self.match('\n')
        self.newline()
        self.currentIndent = None
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mACTION(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = ACTION
        _saveIndex = 0
        startCol = self.getColumn()
        if (self.LA(1)==u'<') and (self.LA(2)==u'\\') and (_tokenSet_2.member(self.LA(3))) and (_tokenSet_3.member(self.LA(4))) and (True) and (True) and (True):
            pass
            buf = u""
            uc = u"\000"
            _saveIndex = self.text.length()
            self.match('<')
            self.text.setLength(_saveIndex)
            _cnt13= 0
            while True:
                if (self.LA(1)==u'\\'):
                    pass
                    uc=self.mESC_CHAR(False)
                    buf += uc
                else:
                    break
                
                _cnt13 += 1
            if _cnt13 < 1:
                self.raise_NoViableAlt(self.LA(1))
            _saveIndex = self.text.length()
            self.match('>')
            self.text.setLength(_saveIndex)
            self.text.setLength(_begin) ; self.text.append(buf)
            _ttype = LITERAL
        elif (self.LA(1)==u'<') and (self.LA(2)==u'!') and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and (True) and (True) and (True):
            pass
            self.mCOMMENT(False)
            _ttype = SKIP
        elif (self.LA(1)==u'<') and (_tokenSet_4.member(self.LA(2))) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
            pass
            if (self.LA(1)==u'<') and (self.LA(2)==u'i') and (self.LA(3)==u'f') and (self.LA(4)==u' ' or self.LA(4)==u'(') and (_tokenSet_5.member(self.LA(5))) and ((self.LA(6) >= u'\u0001' and self.LA(6) <= u'\ufffe')) and ((self.LA(7) >= u'\u0001' and self.LA(7) <= u'\ufffe')):
                pass
                _saveIndex = self.text.length()
                self.match('<')
                self.text.setLength(_saveIndex)
                self.match("if")
                while True:
                    if (self.LA(1)==u' '):
                        pass
                        _saveIndex = self.text.length()
                        self.match(' ')
                        self.text.setLength(_saveIndex)
                    else:
                        break
                    
                self.match("(")
                self.mIF_EXPR(False)
                self.match(")")
                _saveIndex = self.text.length()
                self.match('>')
                self.text.setLength(_saveIndex)
                _ttype = IF
                if (self.LA(1)==u'\n' or self.LA(1)==u'\r'):
                    pass
                    _saveIndex = self.text.length()
                    self.mNL(False)
                    self.text.setLength(_saveIndex)
                    self.newline()
                else: ## <m4>
                        pass
                    
            elif (self.LA(1)==u'<') and (self.LA(2)==u'e') and (self.LA(3)==u'n') and (self.LA(4)==u'd') and (self.LA(5)==u'i') and (self.LA(6)==u'f') and (self.LA(7)==u'>'):
                pass
                _saveIndex = self.text.length()
                self.match('<')
                self.text.setLength(_saveIndex)
                self.match("endif")
                _saveIndex = self.text.length()
                self.match('>')
                self.text.setLength(_saveIndex)
                _ttype = ENDIF
                if ((self.LA(1)==u'\n' or self.LA(1)==u'\r') and (startCol==1)):
                    pass
                    _saveIndex = self.text.length()
                    self.mNL(False)
                    self.text.setLength(_saveIndex)
                    self.newline()
                else: ## <m4>
                        pass
                    
            elif (self.LA(1)==u'<') and (self.LA(2)==u'e') and (self.LA(3)==u'l') and (self.LA(4)==u's') and (self.LA(5)==u'e') and (self.LA(6)==u'>') and (True):
                pass
                _saveIndex = self.text.length()
                self.match('<')
                self.text.setLength(_saveIndex)
                self.match("else")
                _saveIndex = self.text.length()
                self.match('>')
                self.text.setLength(_saveIndex)
                _ttype = ELSE
                if (self.LA(1)==u'\n' or self.LA(1)==u'\r'):
                    pass
                    _saveIndex = self.text.length()
                    self.mNL(False)
                    self.text.setLength(_saveIndex)
                    self.newline()
                else: ## <m4>
                        pass
                    
            elif (self.LA(1)==u'<') and (self.LA(2)==u'@') and (_tokenSet_6.member(self.LA(3))) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and ((self.LA(5) >= u'\u0001' and self.LA(5) <= u'\ufffe')) and ((self.LA(6) >= u'\u0001' and self.LA(6) <= u'\ufffe')) and (True):
                pass
                _saveIndex = self.text.length()
                self.match('<')
                self.text.setLength(_saveIndex)
                _saveIndex = self.text.length()
                self.match('@')
                self.text.setLength(_saveIndex)
                _cnt22= 0
                while True:
                    if (_tokenSet_6.member(self.LA(1))):
                        pass
                        self.match(_tokenSet_6)
                    else:
                        break
                    
                    _cnt22 += 1
                if _cnt22 < 1:
                    self.raise_NoViableAlt(self.LA(1))
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in u'(':
                    pass
                    _saveIndex = self.text.length()
                    self.match("()")
                    self.text.setLength(_saveIndex)
                    _saveIndex = self.text.length()
                    self.match('>')
                    self.text.setLength(_saveIndex)
                    _ttype = REGION_REF
                elif la1 and la1 in u'>':
                    pass
                    _saveIndex = self.text.length()
                    self.match('>')
                    self.text.setLength(_saveIndex)
                    _ttype = REGION_DEF
                    t = self.text.getString(_begin)
                    self.text.setLength(_begin) ; self.text.append(t+"::=")
                    if (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
                        pass
                        la1 = self.LA(1)
                        if False:
                            pass
                        elif la1 and la1 in u'\r':
                            pass
                            _saveIndex = self.text.length()
                            self.match('\r')
                            self.text.setLength(_saveIndex)
                        elif la1 and la1 in u'\n':
                            pass
                        else:
                                self.raise_NoViableAlt(self.LA(1))
                            
                        _saveIndex = self.text.length()
                        self.match('\n')
                        self.text.setLength(_saveIndex)
                        self.newline()
                    elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True):
                        pass
                    else:
                        self.raise_NoViableAlt(self.LA(1))
                    
                    atLeft = False
                    _cnt29= 0
                    while True:
                        if (((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True) and (not (self.upcomingAtEND(1) or (self.upcomingNewline(1) and self.upcomingAtEND(2))))):
                            pass
                            if (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True):
                                pass
                                la1 = self.LA(1)
                                if False:
                                    pass
                                elif la1 and la1 in u'\r':
                                    pass
                                    self.match('\r')
                                elif la1 and la1 in u'\n':
                                    pass
                                else:
                                        self.raise_NoViableAlt(self.LA(1))
                                    
                                self.match('\n')
                                self.newline()
                                atLeft = True
                            elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True):
                                pass
                                self.matchNot(antlr.EOF_CHAR)
                                atLeft = False
                            else:
                                self.raise_NoViableAlt(self.LA(1))
                            
                        else:
                            break
                        
                        _cnt29 += 1
                    if _cnt29 < 1:
                        self.raise_NoViableAlt(self.LA(1))
                    if (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True):
                        pass
                        la1 = self.LA(1)
                        if False:
                            pass
                        elif la1 and la1 in u'\r':
                            pass
                            _saveIndex = self.text.length()
                            self.match('\r')
                            self.text.setLength(_saveIndex)
                        elif la1 and la1 in u'\n':
                            pass
                        else:
                                self.raise_NoViableAlt(self.LA(1))
                            
                        _saveIndex = self.text.length()
                        self.match('\n')
                        self.text.setLength(_saveIndex)
                        self.newline()
                        atLeft = True
                    elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and (True) and (True) and (True) and (True) and (True) and (True):
                        pass
                    else:
                        self.raise_NoViableAlt(self.LA(1))
                    
                    if (self.LA(1)==u'<') and (self.LA(2)==u'@'):
                        pass
                        _saveIndex = self.text.length()
                        self.match("<@end>")
                        self.text.setLength(_saveIndex)
                    elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and (True):
                        pass
                        self.matchNot(antlr.EOF_CHAR)
                        self.this.error("missing region "+t+" <@end> tag")
                    else:
                        self.raise_NoViableAlt(self.LA(1))
                    
                    if ((self.LA(1)==u'\n' or self.LA(1)==u'\r') and (atLeft)):
                        pass
                        la1 = self.LA(1)
                        if False:
                            pass
                        elif la1 and la1 in u'\r':
                            pass
                            _saveIndex = self.text.length()
                            self.match('\r')
                            self.text.setLength(_saveIndex)
                        elif la1 and la1 in u'\n':
                            pass
                        else:
                                self.raise_NoViableAlt(self.LA(1))
                            
                        _saveIndex = self.text.length()
                        self.match('\n')
                        self.text.setLength(_saveIndex)
                        self.newline()
                    else: ## <m4>
                            pass
                        
                else:
                        self.raise_NoViableAlt(self.LA(1))
                    
            elif (self.LA(1)==u'<') and (_tokenSet_4.member(self.LA(2))) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
                pass
                _saveIndex = self.text.length()
                self.match('<')
                self.text.setLength(_saveIndex)
                self.mEXPR(False)
                _saveIndex = self.text.length()
                self.match('>')
                self.text.setLength(_saveIndex)
            else:
                self.raise_NoViableAlt(self.LA(1))
            
            t = ChunkToken(_ttype, self.text.getString(_begin), self.currentIndent)
            _token = t
        else:
            self.raise_NoViableAlt(self.LA(1))
        
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mESC_CHAR(self, _createToken):    
        uc='\u0000'
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = ESC_CHAR
        _saveIndex = 0
        a = None
        b = None
        c = None
        d = None
        if (self.LA(1)==u'\\') and (self.LA(2)==u'n'):
            pass
            _saveIndex = self.text.length()
            self.match("\\n")
            self.text.setLength(_saveIndex)
            uc = '\n'
        elif (self.LA(1)==u'\\') and (self.LA(2)==u'r'):
            pass
            _saveIndex = self.text.length()
            self.match("\\r")
            self.text.setLength(_saveIndex)
            uc = '\r'
        elif (self.LA(1)==u'\\') and (self.LA(2)==u't'):
            pass
            _saveIndex = self.text.length()
            self.match("\\t")
            self.text.setLength(_saveIndex)
            uc = '\t'
        elif (self.LA(1)==u'\\') and (self.LA(2)==u' '):
            pass
            _saveIndex = self.text.length()
            self.match("\\ ")
            self.text.setLength(_saveIndex)
            uc = ' '
        elif (self.LA(1)==u'\\') and (self.LA(2)==u'u'):
            pass
            _saveIndex = self.text.length()
            self.match("\\u")
            self.text.setLength(_saveIndex)
            _saveIndex = self.text.length()
            self.mHEX(True)
            self.text.setLength(_saveIndex)
            a = self._returnToken
            _saveIndex = self.text.length()
            self.mHEX(True)
            self.text.setLength(_saveIndex)
            b = self._returnToken
            _saveIndex = self.text.length()
            self.mHEX(True)
            self.text.setLength(_saveIndex)
            c = self._returnToken
            _saveIndex = self.text.length()
            self.mHEX(True)
            self.text.setLength(_saveIndex)
            d = self._returnToken
            uc = unichr(int(a.getText()+b.getText()+c.getText()+d.getText(), 16))
        else:
            self.raise_NoViableAlt(self.LA(1))
        
        self.set_return_token(_createToken, _token, _ttype, _begin)
        return uc
    
    def mCOMMENT(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = COMMENT
        _saveIndex = 0
        startCol = self.getColumn()
        pass
        self.match("<!")
        while True:
            ###  nongreedy exit test
            if ((self.LA(1)==u'!') and (self.LA(2)==u'>') and (True) and (True) and (True) and (True) and (True)):
                break
            if (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
                pass
                self.mNL(False)
                self.newline()
            elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
                pass
                self.matchNot(antlr.EOF_CHAR)
            else:
                break
            
        self.match("!>")
        if ((self.LA(1)==u'\n' or self.LA(1)==u'\r') and ( startCol == 1 )):
            pass
            self.mNL(False)
            self.newline()
        else: ## <m4>
                pass
            
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mIF_EXPR(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = IF_EXPR
        _saveIndex = 0
        pass
        _cnt54= 0
        while True:
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in u'\\':
                pass
                self.mESC(False)
            elif la1 and la1 in u'\n\r':
                pass
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in u'\r':
                    pass
                    self.match('\r')
                elif la1 and la1 in u'\n':
                    pass
                else:
                        self.raise_NoViableAlt(self.LA(1))
                    
                self.match('\n')
                self.newline()
            elif la1 and la1 in u'{':
                pass
                self.mSUBTEMPLATE(False)
            elif la1 and la1 in u'(':
                pass
                self.mNESTED_PARENS(False)
            else:
                if (_tokenSet_7.member(self.LA(1))):
                    pass
                    self.matchNot(')')
                else:
                    break
                
            _cnt54 += 1
        if _cnt54 < 1:
            self.raise_NoViableAlt(self.LA(1))
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mNL(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = NL
        _saveIndex = 0
        if (self.LA(1)==u'\r') and (self.LA(2)==u'\n') and (True) and (True) and (True) and (True) and (True):
            pass
            self.match('\r')
            self.match('\n')
        elif (self.LA(1)==u'\r') and (True) and (True) and (True) and (True) and (True) and (True):
            pass
            self.match('\r')
        elif (self.LA(1)==u'\n'):
            pass
            self.match('\n')
        else:
            self.raise_NoViableAlt(self.LA(1))
        
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mEXPR(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = EXPR
        _saveIndex = 0
        pass
        _cnt42= 0
        while True:
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in u'\\':
                pass
                self.mESC(False)
            elif la1 and la1 in u'\n\r':
                pass
                self.mNL(False)
                self.newline()
            elif la1 and la1 in u'{':
                pass
                self.mSUBTEMPLATE(False)
            else:
                if (self.LA(1)==u'+' or self.LA(1)==u'=') and (self.LA(2)==u'"' or self.LA(2)==u'<'):
                    pass
                    la1 = self.LA(1)
                    if False:
                        pass
                    elif la1 and la1 in u'=':
                        pass
                        self.match('=')
                    elif la1 and la1 in u'+':
                        pass
                        self.match('+')
                    else:
                            self.raise_NoViableAlt(self.LA(1))
                        
                    self.mTEMPLATE(False)
                elif (self.LA(1)==u'+' or self.LA(1)==u'=') and (self.LA(2)==u'{'):
                    pass
                    la1 = self.LA(1)
                    if False:
                        pass
                    elif la1 and la1 in u'=':
                        pass
                        self.match('=')
                    elif la1 and la1 in u'+':
                        pass
                        self.match('+')
                    else:
                            self.raise_NoViableAlt(self.LA(1))
                        
                    self.mSUBTEMPLATE(False)
                elif (self.LA(1)==u'+' or self.LA(1)==u'=') and (_tokenSet_8.member(self.LA(2))):
                    pass
                    la1 = self.LA(1)
                    if False:
                        pass
                    elif la1 and la1 in u'=':
                        pass
                        self.match('=')
                    elif la1 and la1 in u'+':
                        pass
                        self.match('+')
                    else:
                            self.raise_NoViableAlt(self.LA(1))
                        
                    self.match(_tokenSet_8)
                elif (_tokenSet_9.member(self.LA(1))):
                    pass
                    self.matchNot('>')
                else:
                    break
                
            _cnt42 += 1
        if _cnt42 < 1:
            self.raise_NoViableAlt(self.LA(1))
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mESC(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = ESC
        _saveIndex = 0
        pass
        self.match('\\')
        self.matchNot(antlr.EOF_CHAR)
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mSUBTEMPLATE(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = SUBTEMPLATE
        _saveIndex = 0
        pass
        self.match('{')
        while True:
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in u'{':
                pass
                self.mSUBTEMPLATE(False)
            elif la1 and la1 in u'\\':
                pass
                self.mESC(False)
            else:
                if (_tokenSet_10.member(self.LA(1))):
                    pass
                    self.matchNot('}')
                else:
                    break
                
        self.match('}')
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mTEMPLATE(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = TEMPLATE
        _saveIndex = 0
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in u'"':
            pass
            self.match('"')
            while True:
                if (self.LA(1)==u'\\'):
                    pass
                    self.mESC(False)
                elif (_tokenSet_11.member(self.LA(1))):
                    pass
                    self.matchNot('"')
                else:
                    break
                
            self.match('"')
        elif la1 and la1 in u'<':
            pass
            self.match("<<")
            if (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and (True) and (True) and (True):
                pass
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in u'\r':
                    pass
                    _saveIndex = self.text.length()
                    self.match('\r')
                    self.text.setLength(_saveIndex)
                elif la1 and la1 in u'\n':
                    pass
                else:
                        self.raise_NoViableAlt(self.LA(1))
                    
                _saveIndex = self.text.length()
                self.match('\n')
                self.text.setLength(_saveIndex)
                self.newline()
            elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True):
                pass
            else:
                self.raise_NoViableAlt(self.LA(1))
            
            while True:
                ###  nongreedy exit test
                if ((self.LA(1)==u'>') and (self.LA(2)==u'>') and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and (True) and (True) and (True) and (True)):
                    break
                if ((self.LA(1)==u'\r') and (self.LA(2)==u'\n') and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and ((self.LA(5) >= u'\u0001' and self.LA(5) <= u'\ufffe')) and (True) and (True) and ( self.LA(3) == '>' and self.LA(4) == '>' )):
                    pass
                    _saveIndex = self.text.length()
                    self.match('\r')
                    self.text.setLength(_saveIndex)
                    _saveIndex = self.text.length()
                    self.match('\n')
                    self.text.setLength(_saveIndex)
                    self.newline()
                elif ((self.LA(1)==u'\n') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and (True) and (True) and (True) and ( self.LA(2) == '>' and self.LA(3) == '>' )):
                    pass
                    _saveIndex = self.text.length()
                    self.match('\n')
                    self.text.setLength(_saveIndex)
                    self.newline()
                elif (self.LA(1)==u'\n' or self.LA(1)==u'\r') and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and (True) and (True) and (True):
                    pass
                    la1 = self.LA(1)
                    if False:
                        pass
                    elif la1 and la1 in u'\r':
                        pass
                        self.match('\r')
                    elif la1 and la1 in u'\n':
                        pass
                    else:
                            self.raise_NoViableAlt(self.LA(1))
                        
                    self.match('\n')
                    self.newline()
                elif ((self.LA(1) >= u'\u0001' and self.LA(1) <= u'\ufffe')) and ((self.LA(2) >= u'\u0001' and self.LA(2) <= u'\ufffe')) and ((self.LA(3) >= u'\u0001' and self.LA(3) <= u'\ufffe')) and ((self.LA(4) >= u'\u0001' and self.LA(4) <= u'\ufffe')) and (True) and (True) and (True):
                    pass
                    self.matchNot(antlr.EOF_CHAR)
                else:
                    break
                
            self.match(">>")
        else:
                self.raise_NoViableAlt(self.LA(1))
            
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mNESTED_PARENS(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = NESTED_PARENS
        _saveIndex = 0
        pass
        self.match('(')
        _cnt63= 0
        while True:
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in u'(':
                pass
                self.mNESTED_PARENS(False)
            elif la1 and la1 in u'\\':
                pass
                self.mESC(False)
            else:
                if (_tokenSet_12.member(self.LA(1))):
                    pass
                    self.matchNot(')')
                else:
                    break
                
            _cnt63 += 1
        if _cnt63 < 1:
            self.raise_NoViableAlt(self.LA(1))
        self.match(')')
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    def mHEX(self, _createToken):    
        _ttype = 0
        _token = None
        _begin = self.text.length()
        _ttype = HEX
        _saveIndex = 0
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in u'0123456789':
            pass
            self.matchRange(u'0', u'9')
        elif la1 and la1 in u'ABCDEF':
            pass
            self.matchRange(u'A', u'F')
        elif la1 and la1 in u'abcdef':
            pass
            self.matchRange(u'a', u'f')
        else:
                self.raise_NoViableAlt(self.LA(1))
            
        self.set_return_token(_createToken, _token, _ttype, _begin)
    
    

### generate bit set
def mk_tokenSet_0(): 
    data = [0L] * 2048 ### init list
    data[0] =-1152921504606856194L
    for x in xrange(1, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_0 = antlr.BitSet(mk_tokenSet_0())

### generate bit set
def mk_tokenSet_1(): 
    data = [0L] * 2048 ### init list
    data[0] =-5764607523034234882L
    for x in xrange(1, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_1 = antlr.BitSet(mk_tokenSet_1())

### generate bit set
def mk_tokenSet_2(): 
    data = [0L] * 1025 ### init list
    data[0] =4294967296L
    data[1] =14707067533131776L
    return data
_tokenSet_2 = antlr.BitSet(mk_tokenSet_2())

### generate bit set
def mk_tokenSet_3(): 
    data = [0L] * 1025 ### init list
    data[0] =4899634919602388992L
    data[1] =541434314878L
    return data
_tokenSet_3 = antlr.BitSet(mk_tokenSet_3())

### generate bit set
def mk_tokenSet_4(): 
    data = [0L] * 2048 ### init list
    data[0] =-4611686018427387906L
    for x in xrange(1, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_4 = antlr.BitSet(mk_tokenSet_4())

### generate bit set
def mk_tokenSet_5(): 
    data = [0L] * 2048 ### init list
    data[0] =-2199023255554L
    for x in xrange(1, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_5 = antlr.BitSet(mk_tokenSet_5())

### generate bit set
def mk_tokenSet_6(): 
    data = [0L] * 2048 ### init list
    data[0] =-4611687117939015682L
    for x in xrange(1, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_6 = antlr.BitSet(mk_tokenSet_6())

### generate bit set
def mk_tokenSet_7(): 
    data = [0L] * 2048 ### init list
    data[0] =-3298534892546L
    data[1] =-576460752571858945L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_7 = antlr.BitSet(mk_tokenSet_7())

### generate bit set
def mk_tokenSet_8(): 
    data = [0L] * 2048 ### init list
    data[0] =-1152921521786716162L
    data[1] =-576460752303423489L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_8 = antlr.BitSet(mk_tokenSet_8())

### generate bit set
def mk_tokenSet_9(): 
    data = [0L] * 2048 ### init list
    data[0] =-6917537823734113282L
    data[1] =-576460752571858945L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_9 = antlr.BitSet(mk_tokenSet_9())

### generate bit set
def mk_tokenSet_10(): 
    data = [0L] * 2048 ### init list
    data[0] =-2L
    data[1] =-2882303761785552897L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_10 = antlr.BitSet(mk_tokenSet_10())

### generate bit set
def mk_tokenSet_11(): 
    data = [0L] * 2048 ### init list
    data[0] =-17179869186L
    data[1] =-268435457L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_11 = antlr.BitSet(mk_tokenSet_11())

### generate bit set
def mk_tokenSet_12(): 
    data = [0L] * 2048 ### init list
    data[0] =-3298534883330L
    data[1] =-268435457L
    for x in xrange(2, 1023):
        data[x] = -1L
    data[1023] =9223372036854775807L
    return data
_tokenSet_12 = antlr.BitSet(mk_tokenSet_12())
    
### __main__ header action >>> 
if __name__ == '__main__' :
    import sys
    import antlr
    import AngleBracketTemplateLexer
    
    ### create lexer - shall read from stdin
    try:
        for token in AngleBracketTemplateLexer.Lexer():
            print token
            
    except antlr.TokenStreamException, e:
        print "error: exception caught while lexing: ", e
### __main__ header action <<< 




import antlr

class StringTemplateAST(antlr.CommonAST):

    def __init__(self, type=None, text=None):
        super(StringTemplateAST, self).__init__() 

        if type is not None:
            self.setType(type)

        if text is not None:
            self.setText(text)

	# track template for ANONYMOUS blocks
        self.st = None

    def getStringTemplate(self):
        return self.st

    def setStringTemplate(self, st):
        self.st = st




from stringtemplate3.language.Expr import Expr

class StringRef(Expr):
    """
    Represents a chunk of just simple text to spit out; nothing to "evaluate"
    """

    def __init__(self, enclosingTemplate, text):
        super(StringRef, self).__init__(enclosingTemplate)
        self.str = text


    def write(self, this, out):
        """
        Just print out the string; no reference to self because this
        is a literal -- not sensitive to attribute values. These strings
        never wrap because they are not part of an <...> expression.
        <"foo"; wrap="\n"> should wrap though if necessary.
        """
        
        if self.str is not None:
            return out.write(self.str)

        return 0

    
    def __str__(self):
        if self.str:
            return self.str
        return ''




# [The "BSD licence"]
# Copyright (c) 2003-2006 Terence Parr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from stringtemplate3.utils import deprecated


class AttributeRenderer(object):
    """
    This interface describes an object that knows how to format or otherwise
    render an object appropriately.  Usually this is used for locale changes
    for objects such as Date and floating point numbers...  You can either
    have an object that is sensitive to the locale or have a different object
    per locale.

    Each template may have a renderer for each object type or can default
    to the group's renderer or the super group's renderer if the group doesn't
    have one.

    The toString(Object,String) method is used when the user uses the
    format option: $o; format="f"$.  It checks the formatName and applies the
    appropriate formatting.  If the format string passed to the renderer is
    not recognized then simply call toString().
    """

    def __init__(self):
        pass

    def toString(self, o, formatName=None):
        pass


class StringTemplateWriter(object):
    """
    Generic StringTemplate output writer filter.
 
    Literals and the elements of expressions are emitted via write().
    Separators are emitted via writeSeparator() because they must be
    handled specially when wrapping lines (we don't want to wrap
    in between an element and it's separator).
    """

    NO_WRAP = -1
    
    def __init__(self):
        pass

    def pushIndentation(self, indent):
        raise NotImplementedError

    def popIndentation(self):
        raise NotImplementedError

    def pushAnchorPoint(self):
        raise NotImplementedError

    
    def popAnchorPoint(self):
        raise NotImplementedError


    def setLineWidth(self, lineWidth):
        raise NotImplementedError
    

    def write(self, str, wrap=None):
        """
        Write the string and return how many actual chars were written.
        With autoindentation and wrapping, more chars than length(str)
        can be emitted. No wrapping is done unless wrap is not None
        """

        raise NotImplementedError


    def writeWrapSeparator(self, wrap):
        """
        Because we might need to wrap at a non-atomic string boundary
 	(such as when we wrap in between template applications
 	 <data:{v|[<v>]}; wrap>) we need to expose the wrap string
 	writing just like for the separator.
 	"""

        raise NotImplementedError


    def writeSeparator(self, str):
        """
        Write a separator.  Same as write() except that a \n cannot
        be inserted before emitting a separator.
 	"""

        raise NotImplementedError


class AutoIndentWriter(StringTemplateWriter):
    """
    Essentially a char filter that knows how to auto-indent output
    by maintaining a stack of indent levels.  I set a flag upon newline
    and then next nonwhitespace char resets flag and spits out indention.
    The indent stack is a stack of strings so we can repeat original indent
    not just the same number of columns (don't have to worry about tabs vs
    spaces then).

    Anchors are char positions (tabs won't work) that indicate where all
    future wraps should justify to.  The wrap position is actually the
    larger of either the last anchor or the indentation level.
    
    This is a filter on a Writer.

    It may be screwed up for '\r' '\n' on PC's.
    """

    def __init__(self, out):
        StringTemplateWriter.__init__(self)

	## stack of indents
        self.indents = [None] # start with no indent

        ## Stack of integer anchors (char positions in line)
        self.anchors = []

        self.out = out
        self.atStartOfLine = True

        ## Track char position in the line (later we can think about tabs).
        #  Indexed from 0.  We want to keep charPosition <= lineWidth.
        #  This is the position we are *about* to write not the position
        #  last written to.
        self.charPosition = 0
        self.lineWidth = self.NO_WRAP

        self.charPositionOfStartOfExpr = 0


    @deprecated
    def setLineWidth(self, lineWidth):
        self.lineWidth = lineWidth

        
    def pushIndentation(self, indent):
        """
        Push even blank (null) indents as they are like scopes; must
        be able to pop them back off stack.
        """

        self.indents.append(indent)


    def popIndentation(self):
        return self.indents.pop(-1)


    def getIndentationWidth(self):
        return sum(len(ind) for ind in self.indents if ind is not None)


    def pushAnchorPoint(self):
        self.anchors.append(self.charPosition)


    def popAnchorPoint(self):
        self.anchors.pop(-1)


    def write(self, text, wrap=None):
        """
        Write out a string literal or attribute expression or expression
        element.

        If doing line wrap, then check wrap before emitting this str.  If
        at or beyond desired line width then emit a \n and any indentation
        before spitting out this str.
        """

        assert isinstance(text, basestring), repr(text)
        
        n = 0
        if wrap is not None:
            n += self.writeWrapSeparator(wrap)
            
        for c in text:
            if c == '\n':
                self.atStartOfLine = True
                self.charPosition = -1 # set so the write below sets to 0

            else:
                if self.atStartOfLine:
                    n += self.indent()
                    self.atStartOfLine = False

            n += 1
            self.out.write(c)
            self.charPosition += 1

        return n
        

    def writeWrapSeparator(self, wrap):
        n = 0
        
        # if want wrap and not already at start of line (last char was \n)
        # and we have hit or exceeded the threshold
        if ( self.lineWidth != self.NO_WRAP and
             not self.atStartOfLine and
             self.charPosition >= self.lineWidth ):
            # ok to wrap
            # Walk wrap string and look for A\nB.  Spit out A\n
            # then spit indent or anchor, whichever is larger
            # then spit out B
            for c in wrap:
                if c == '\n':
                    n += 1
                    self.out.write(c)
                    #atStartOfLine = true;
                    self.charPosition = 0
                    
                    indentWidth = self.getIndentationWidth()
                    try:
                        lastAnchor = self.anchors[-1]
                    except IndexError: # no anchors saved
                        lastAnchor = 0
                        
                    if lastAnchor > indentWidth:
                        # use anchor not indentation
                        n += self.indent(lastAnchor)

                    else:
                        # indent is farther over than last anchor, ignore anchor
                        n += self.indent()

                    # continue writing any chars out

                else: # write A or B part
                    n += 1
                    self.out.write(c)
                    self.charPosition += 1

        return n

    def writeSeparator(self, text):
        return self.write(text)


    def indent(self, spaces=None):
        if spaces is None:
            n = 0
            for ind in self.indents:
                if ind is not None:
                    n += len(ind)
                    self.out.write(ind)

            self.charPosition += n
            return n

        else:
            self.out.write(' ' * spaces)
            self.charPosition += spaces
            return spaces


## Just pass through the text
#
class NoIndentWriter(AutoIndentWriter):

    def __init__(self, out):
        super(NoIndentWriter, self).__init__(out)


    def write(self, str, wrap=None):
        self.out.write(str)
        return len(str)




### $ANTLR 2.7.7 (2006-11-01): "interface.g" -> "InterfaceParser.py"$
### import antlr and other modules ..
import sys
import antlr

version = sys.version.split()[0]
if version < '2.2.1':
    False = 0
if version < '2.3':
    True = not False
### header action >>> 
#
# [The "BSD licence"]
# Copyright (c) 2003-2004 Terence Parr
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 


import sys
import traceback

from stringtemplate3.language.FormalArgument import FormalArgument
### header action <<< 
### preamble action>>>

### preamble action <<<

### import antlr.Token 
from antlr import Token
### >>>The Known Token Types <<<
SKIP                = antlr.SKIP
INVALID_TYPE        = antlr.INVALID_TYPE
EOF_TYPE            = antlr.EOF_TYPE
EOF                 = antlr.EOF
NULL_TREE_LOOKAHEAD = antlr.NULL_TREE_LOOKAHEAD
MIN_USER_TYPE       = antlr.MIN_USER_TYPE
LITERAL_interface = 4
ID = 5
SEMI = 6
LITERAL_optional = 7
LPAREN = 8
RPAREN = 9
COMMA = 10
COLON = 11
SL_COMMENT = 12
ML_COMMENT = 13
WS = 14


###/** Match an ST group interface.  Just a list of template names with args.
### *  Here is a sample interface file:
### *
### *	interface nfa;
### *	nfa(states,edges);
### *	optional state(name);
### */
class Parser(antlr.LLkParser):
    ### user action >>>
    def reportError(self, e):
       if self.group_:
           self.group_.error("template group interface parse error", e)
       else:
           sys.stderr.write("template group interface parse error: " + str(e) + '\n')
           traceback.print_exc()
    ### user action <<<
    
    def __init__(self, *args, **kwargs):
        antlr.LLkParser.__init__(self, *args, **kwargs)
        self.tokenNames = _tokenNames
        ### __init__ header action >>> 
        self.groupI = None
        ### __init__ header action <<< 
        
    def groupInterface(self,
        groupI
    ):    
        
        name = None
        self.groupI = groupI
        try:      ## for error handling
            pass
            self.match(LITERAL_interface)
            name = self.LT(1)
            self.match(ID)
            groupI.name = name.getText()
            self.match(SEMI)
            _cnt3= 0
            while True:
                if (self.LA(1)==ID or self.LA(1)==LITERAL_optional):
                    pass
                    self.template(groupI)
                else:
                    break
                
                _cnt3 += 1
            if _cnt3 < 1:
                raise antlr.NoViableAltException(self.LT(1), self.getFilename())
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            self.consume()
            self.consumeUntil(_tokenSet_0)
        
    
    def template(self,
        groupI
    ):    
        
        opt = None
        name = None
        formalArgs = {} # leave blank if no args
        templateName = None
        try:      ## for error handling
            pass
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in [LITERAL_optional]:
                pass
                opt = self.LT(1)
                self.match(LITERAL_optional)
            elif la1 and la1 in [ID]:
                pass
            else:
                    raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                
            name = self.LT(1)
            self.match(ID)
            self.match(LPAREN)
            la1 = self.LA(1)
            if False:
                pass
            elif la1 and la1 in [ID]:
                pass
                formalArgs=self.args()
            elif la1 and la1 in [RPAREN]:
                pass
            else:
                    raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                
            self.match(RPAREN)
            self.match(SEMI)
            templateName = name.getText()
            groupI.defineTemplate(templateName, formalArgs, opt != None)
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            self.consume()
            self.consumeUntil(_tokenSet_1)
        
    
    def args(self):    
        args={}
        
        a = None
        b = None
        try:      ## for error handling
            pass
            a = self.LT(1)
            self.match(ID)
            args[a.getText()] = FormalArgument(a.getText())
            while True:
                if (self.LA(1)==COMMA):
                    pass
                    self.match(COMMA)
                    b = self.LT(1)
                    self.match(ID)
                    args[b.getText()] = FormalArgument(b.getText())
                else:
                    break
                
        
        except antlr.RecognitionException, ex:
            self.reportError(ex)
            self.consume()
            self.consumeUntil(_tokenSet_2)
        
        return args
    

_tokenNames = [
    "<0>", 
    "EOF", 
    "<2>", 
    "NULL_TREE_LOOKAHEAD", 
    "\"interface\"", 
    "ID", 
    "SEMI", 
    "\"optional\"", 
    "LPAREN", 
    "RPAREN", 
    "COMMA", 
    "COLON", 
    "SL_COMMENT", 
    "ML_COMMENT", 
    "WS"
]
    

### generate bit set
def mk_tokenSet_0(): 
    ### var1
    data = [ 2L, 0L]
    return data
_tokenSet_0 = antlr.BitSet(mk_tokenSet_0())

### generate bit set
def mk_tokenSet_1(): 
    ### var1
    data = [ 162L, 0L]
    return data
_tokenSet_1 = antlr.BitSet(mk_tokenSet_1())

### generate bit set
def mk_tokenSet_2(): 
    ### var1
    data = [ 512L, 0L]
    return data
_tokenSet_2 = antlr.BitSet(mk_tokenSet_2())
    
